from typing import Optional, List, Any

from boto3.dynamodb.conditions import Key
from boto3.exceptions import ResourceNotExistsError
from pydantic import BaseModel

from inoft_vocal_engine.databases.dynamodb.dynamodb_core import DynamoDbCoreAdapter, GlobalSecondaryIndex, PrimaryIndex, \
    Response, DynamoDBMapObjectSetter
from inoft_vocal_engine.databases.dynamodb.dynamodb_utils import Utils
from inoft_vocal_engine.import_integrations.botpress.engine_models import ProcessedNodeElement, NodeDatabaseItem, \
    NodeElementDataReceivedFromUser, DatabaseTransitionConditionItem
from inoft_vocal_engine.models.actions import BaseAction


class ProjectsDiagramsDataDynamoDbClient(DynamoDbCoreAdapter):
    def __init__(self, table_name: str, region_name: str):
        primary_index = PrimaryIndex(hash_key_name="nodeId", hash_key_variable_python_type=str)

        # todo: setup the global secondary indexes (they are currently a copy of the text contents database client)
        globals_secondary_indexes = [
            GlobalSecondaryIndex(hash_key_name="accountId", hash_key_variable_python_type=str, projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="accountProjectId", hash_key_variable_python_type=str, projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="accountProjectInstanceId", hash_key_variable_python_type=str, projection_type="ALL")
        ]
        super().__init__(table_name=table_name, region_name=region_name, primary_index=primary_index,
                         create_table=True, global_secondary_indexes=globals_secondary_indexes)

    def put_new_node(self, node_item: NodeDatabaseItem):
        self._put_item_dict(item_dict=node_item.dict())

    def _query_nodes_by_key(self, key_name: str, key_value: str) -> Response:
        return self._query_by_key(key_name=key_name, index_name=key_name, key_value=key_value)

    def query_nodes_by_instance_id(self, instance_id: str):
        response = self._query_nodes_by_key(key_name='accountProjectInstanceId', key_value=instance_id)
        return response

    def query_nodes_by_project_id(self, project_id: str) -> Response:
        response = self._query_nodes_by_key(key_name='accountProjectId', key_value=project_id)
        return response

    def _get_nodes_by_key(self, key_name: str, key_value: str) -> List[ProcessedNodeElement]:
        response = self._query_nodes_by_key(key_name=key_name, key_value=key_value)
        output_nodes: List[ProcessedNodeElement] = list()
        for item in response.items:
            if isinstance(item, dict):
                current_node_database_item = NodeDatabaseItem(**item)
                output_nodes.append(current_node_database_item.to_processed_node_element())
        return output_nodes

    def get_nodes_by_instance_id(self, instance_id: str) -> List[ProcessedNodeElement]:
        nodes = self._get_nodes_by_key(key_name='accountProjectInstanceId', key_value=instance_id)
        return nodes

    def get_nodes_by_project_id(self, project_id: str) -> List[ProcessedNodeElement]:
        nodes = self._get_nodes_by_key(key_name='accountProjectId', key_value=project_id)
        return nodes

    # todo: finish the client (update, get by latest, get by filter, and delete node)

    def update_node(self, node_element_data: NodeElementDataReceivedFromUser) -> Response:
        objects_setters: List[DynamoDBMapObjectSetter] = list()

        if node_element_data.nodeName is not None:
            objects_setters.append(DynamoDBMapObjectSetter(object_path="nodeName", value_to_set=node_element_data.nodeName))
        if node_element_data.nodeType is not None:
            objects_setters.append(DynamoDBMapObjectSetter(object_path="nodeType", value_to_set=node_element_data.nodeType))
        if node_element_data.actions is not None:
            objects_setters.append(DynamoDBMapObjectSetter(object_path="actions", value_to_set=node_element_data.actions))
        if node_element_data.transitions is not None:
            objects_setters.append(DynamoDBMapObjectSetter(object_path="transitions", value_to_set=node_element_data.transitions))
        if node_element_data.positionY is not None:
            objects_setters.append(DynamoDBMapObjectSetter(object_path="positionX", value_to_set=node_element_data.positionX))
        if node_element_data.positionY is not None:
            objects_setters.append(DynamoDBMapObjectSetter(object_path="positionY", value_to_set=node_element_data.positionY))
        if node_element_data.temp_on_receive is not None:
            objects_setters.append(DynamoDBMapObjectSetter(object_path="temp_on_receive", value_to_set=node_element_data.temp_on_receive))

        response = self._add_update_multiple_data_elements_to_map(
            key_name="nodeId", key_value=node_element_data.nodeId, objects_setters=objects_setters
        )
        return response

    def add_update_transition_to_node(self, transition_condition: DatabaseTransitionConditionItem) -> Response:
        response = self._add_update_data_element_to_map(
            key_name="nodeId", key_value=transition_condition.sourceNodeId,
            object_path_elements={"transitions": dict, transition_condition.transitionId: dict},
            element_values=transition_condition.to_database()
        )
        return response

    def add_update_action_to_node(self, action_instance: BaseAction) -> Response:
        # The real action_instance will not be of type BaseAction, but should always have BaseAction as of one of its parent.
        response = self._add_update_data_element_to_map(
            key_name="nodeId", key_value=action_instance.nodeId,
            object_path_elements={"actions": dict, action_instance.actionId: dict},
            element_values=action_instance.to_database()
        )
        return response


    """
    def get_latest_created(self, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        return self._get_latest(index_name="stateId-creationTimestamp",
                                num_latest_items=num_latest_items, exclusive_start_key=exclusive_start_key)

    def get_latest_updated(self, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        return self._get_latest(index_name="stateId-lastModificationTimestamp",
                                num_latest_items=num_latest_items, exclusive_start_key=exclusive_start_key)

    def get_by_id(self, account_project_id: str) -> (dict, bool):
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.get_item(TableName=self.table_name, Key={"elementId": account_project_id}, ConsistentRead=True)
            if "Item" in response:
                return Utils.dynamodb_to_python(dynamodb_object=response["Item"]), True
            else:
                return dict(), False
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")

    def get_by_character_name(self, character_name: str) -> Response:
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.scan(FilterExpression=Attr('characterNames').contains(character_name))
            if "Items" in response:
                return Response(Utils.dynamodb_to_python(dynamodb_object=response))
            else:
                return Response({})
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")

    def get_by_text_search(self, text_to_search: str) -> Response:
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.scan(FilterExpression=Attr('crudeText').contains(text_to_search))
            if "Items" in response:
                return Response(Utils.dynamodb_to_python(dynamodb_object=response))
            else:
                return Response({})
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")
    """




import os
from typing import Optional

from inoft_vocal_engine.audacity.pipeclient import PipeClient


class AudacityClient:
    EXPORT_TYPE_MP3 = "mp3"
    EXPORT_TYPE_WAV = "wav"
    EXPORT_MODE_ALL = "All"
    EXPORT_MODE_SELECTION = "Selection"

    def __init__(self):
        self.client = PipeClient()

    def set_clip(self, clip_id: int, track_number: Optional[int] = None, seconds_start: Optional[float] = None):
        command = f"SetClip: Clip={clip_id}"
        if track_number is not None:
            command += f" At={track_number}"
        if seconds_start is not None:
            command += f" Start={seconds_start}"
        self.client.write(command)

    def select_all(self):
        self.client.write("SelectAll:")

    def select_next_clip(self):
        self.client.write("SelNextClip:")

    def select_previous_clip(self):
        self.client.write("SelPrevClip:")

    def select_track(self, track_number: int = 0):
        self.client.write(f"SelectTracks: Track={track_number} Mode=Add")

    def import_file(self, filepath: str, track_number: int = 0):
        self.client.write(f"Select: Track={track_number}")
        self.client.write(f"Import2: Filename={filepath}")
        # self.set_clip(clip_id=0)

    def delete_all_audio(self):
        self.client.write("SelectAll:")
        self.client.write("Cut:")

    def delete_all_tracks(self):
        self.client.write("SelectTracks: Track=First TrackCount=Last Mode=Add")
        # We add all the tracks from the First to the Last to a selection
        self.client.write("RemoveTracks:")
        # Then we remove the tracks in our selection

    def mix_and_render_all_elements(self):
        self.select_all()
        self.mix_and_render_tracks()

    def mix_and_render_tracks(self):
        self.client.write("MixAndRender:")

    def new_project(self):
        self.client.write("New:")

    def save_project(self, project_output_filepath: str, add_to_history: Optional[bool] = True, compress: Optional[bool] = False):
        if os.path.isfile(project_output_filepath):
            # If audacity try to save a project over an existing file, it might crash
            os.remove(project_output_filepath)
        self.client.write(f"SaveProject2: Filename={project_output_filepath} AddToHistory={add_to_history} Compress={compress}")

    def export(self, filepath: str, num_channels: int = 1, export_mode: str = EXPORT_MODE_ALL):
        self.client.write(f"Export2: Filename={filepath} NumChannels={num_channels}")

    def get_info(self):
        print(self.client.write("GetInfo: Type=Tracks"))


if __name__ == "__main__":
    client = AudacityClient()

    client.mix_and_render_all_elements()
    # client.select_track(track_number=3)
    # client.set_clip(clip_id=0, track_number=0, seconds_start=2.0)

    # client.import_file(filepath="C:/Users/LABOURDETTE/Documents/MAGIX/2020-05-20_01_24.wav")  # "F:/test1.mp3")
    # client.save_project(project_output_filepath="F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/project.aup")
    # To note, audacity will only accept wav files, and no spaces can be present in the filepath. So the files must
    # be moved to a temporary folder (if they have spaces) and be converted to wav files if they are not in wav.
    print("done")




from inoft_vocal_engine import Sound

text = """
*Léo* Ont pourrais essayer de récuperer des informations en utilisant leurs radio.
*Willie* Comment ont ferrais, il faudrais qu'ils croient qu'ont soit dans leurs camp ?
*Léo* Je sais pas exactement, ont trouveras. Ont pourrais essayer de capturer un Bosch.
*Willie* Ca pourrais marcher...
"""

if __name__ == "__main__":
    from inoft_vocal_engine.audacity.client import AudacityClient
    from inoft_vocal_engine.speech_synthesis.polly.client import PollyClient
    from inoft_vocal_engine.speech_synthesis.polly import VOICES
    audacity = AudacityClient()
    """polly = PollyClient()

    leo = Sound(polly.synthesize(text="Ont pourrais essayer de récuperer des informations en utilisant leurs radio.",
                           voice_id=VOICES.French_France_Male_MATHIEU.id,
                                 filepath_to_save_to="F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/leo.mp3"))

    willie = Sound(polly.synthesize(text="Comment ont ferrais, il faudrais qu'ils croient qu'ont soit dans leurs camp ?",
                                    voice_id=VOICES.French_France_Female_CELINE.id,
                                    filepath_to_save_to="F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/willie.mp3"))
    """

    leo = Sound("F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/leo.mp3")
    willie = Sound("F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/willie.mp3")

    audacity.delete_all_tracks()
    audacity.import_file(leo.local_filepath, track_number=1)
    # audacity.import_file(willie.local_filepath, track_number=2)
    audacity.set_clip(clip_id=1, track_number=0, seconds_start=leo.duration_seconds)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2019 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import array
import ctypes
import ctypes.util
import logging
import time
import math
import os.path
import sys
import bisect
import threading
import traceback

from collections import deque
from bisect import insort

from . import utils
from .rtp import RTPPacket, SilencePacket, FECPacket
from .errors import DiscordException

log = logging.getLogger(__name__)

c_int_ptr   = ctypes.POINTER(ctypes.c_int)
c_int16_ptr = ctypes.POINTER(ctypes.c_int16)
c_float_ptr = ctypes.POINTER(ctypes.c_float)

_lib = None

class EncoderStruct(ctypes.Structure):
    pass

class DecoderStruct(ctypes.Structure):
    pass

EncoderStructPtr = ctypes.POINTER(EncoderStruct)
DecoderStructPtr = ctypes.POINTER(DecoderStruct)

## Some constants from opus_defines.h
# Error codes
OK      = 0
BAD_ARG = -1

# Encoder CTLs
APPLICATION_AUDIO    = 2049
APPLICATION_VOIP     = 2048
APPLICATION_LOWDELAY = 2051

CTL_SET_BITRATE      = 4002
CTL_SET_BANDWIDTH    = 4008
CTL_SET_FEC          = 4012
CTL_SET_PLP          = 4014
CTL_SET_SIGNAL       = 4024

# Decoder CTLs
CTL_SET_GAIN             = 4034
CTL_LAST_PACKET_DURATION = 4039

band_ctl = {
    'narrow': 1101,
    'medium': 1102,
    'wide': 1103,
    'superwide': 1104,
    'full': 1105,
}

signal_ctl = {
    'auto': -1000,
    'voice': 3001,
    'music': 3002,
}

def _err_lt(result, func, args):
    if result < OK:
        log.info('error has happened in %s', func.__name__)
        raise OpusError(result)
    return result

def _err_ne(result, func, args):
    ret = args[-1]._obj
    if ret.value != OK:
        log.info('error has happened in %s', func.__name__)
        raise OpusError(ret.value)
    return result

# A list of exported functions.
# The first argument is obviously the name.
# The second one are the types of arguments it takes.
# The third is the result type.
# The fourth is the error handler.
exported_functions = [
    ('opus_strerror',
        [ctypes.c_int], ctypes.c_char_p, None),
    ('opus_packet_get_bandwidth',
        [ctypes.c_char_p], ctypes.c_int, _err_lt),
    ('opus_packet_get_nb_channels',
        [ctypes.c_char_p], ctypes.c_int, _err_lt),
    ('opus_packet_get_nb_frames',
        [ctypes.c_char_p, ctypes.c_int], ctypes.c_int, _err_lt),
    ('opus_packet_get_samples_per_frame',
        [ctypes.c_char_p, ctypes.c_int], ctypes.c_int, _err_lt),

    ('opus_encoder_get_size',
        [ctypes.c_int], ctypes.c_int, None),
    ('opus_encoder_create',
        [ctypes.c_int, ctypes.c_int, ctypes.c_int, c_int_ptr], EncoderStructPtr, _err_ne),
    ('opus_encode',
        [EncoderStructPtr, c_int16_ptr, ctypes.c_int, ctypes.c_char_p, ctypes.c_int32], ctypes.c_int32, _err_lt),
    ('opus_encoder_ctl',
        None, ctypes.c_int32, _err_lt),
    ('opus_encoder_destroy',
        [EncoderStructPtr], None, None),

    ('opus_decoder_get_size',
        [ctypes.c_int], ctypes.c_int, None),
    ('opus_decoder_create',
        [ctypes.c_int, ctypes.c_int, c_int_ptr], DecoderStructPtr, _err_ne),
    ('opus_decoder_get_nb_samples',
        [DecoderStructPtr, ctypes.c_char_p, ctypes.c_int32], ctypes.c_int, _err_lt),
    ('opus_decode',
        [DecoderStructPtr, ctypes.c_char_p, ctypes.c_int32, c_int16_ptr, ctypes.c_int, ctypes.c_int],
        ctypes.c_int, _err_lt),
    ('opus_decoder_ctl',
        None, ctypes.c_int32, _err_lt),
    ('opus_decoder_destroy',
        [DecoderStructPtr], None, None)
]

def libopus_loader(name):
    # create the library...
    lib = ctypes.cdll.LoadLibrary(name)

    # register the functions...
    for item in exported_functions:
        func = getattr(lib, item[0])

        try:
            if item[1]:
                func.argtypes = item[1]

            func.restype = item[2]
        except KeyError:
            pass

        try:
            if item[3]:
                func.errcheck = item[3]
        except KeyError:
            log.info("Error assigning check function to %s", item[0])

    return lib

def _load_default():
    global _lib
    try:
        if sys.platform == 'win32':
            _basedir = os.path.dirname(os.path.abspath(__file__))
            _bitness = 'x64' if sys.maxsize > 2**32 else 'x86'
            _filename = os.path.join(_basedir, 'bin', 'libopus-0.{}.dll'.format(_bitness))
            _lib = libopus_loader(_filename)
        else:
            _lib = libopus_loader(ctypes.util.find_library('opus'))
    except Exception as e:
        _lib = None
        log.warning("Unable to load opus lib, %s", e)

    return _lib is not None

def load_opus(name):
    """Loads the libopus shared library for use with voice.

    If this function is not called then the library uses the function
    :func:`ctypes.util.find_library` and then loads that one if available.

    Not loading a library and attempting to use PCM based AudioSources will
    lead to voice not working.

    This function propagates the exceptions thrown.

    .. warning::

        The bitness of the library must match the bitness of your python
        interpreter. If the library is 64-bit then your python interpreter
        must be 64-bit as well. Usually if there's a mismatch in bitness then
        the load will throw an exception.

    .. note::

        On Windows, this function should not need to be called as the binaries
        are automatically loaded.

    .. note::

        On Windows, the .dll extension is not necessary. However, on Linux
        the full extension is required to load the library, e.g. ``libopus.so.1``.
        On Linux however, :func:`ctypes.util.find_library` will usually find the library automatically
        without you having to call this.

    Parameters
    ----------
    name: :class:`str`
        The filename of the shared library.
    """
    global _lib
    _lib = libopus_loader(name)

def is_loaded():
    """Function to check if opus lib is successfully loaded either
    via the :func:`ctypes.util.find_library` call of :func:`load_opus`.

    This must return ``True`` for voice to work.

    Returns
    -------
    :class:`bool`
        Indicates if the opus library has been loaded.
    """
    global _lib
    return _lib is not None

class OpusError(DiscordException):
    """An exception that is thrown for libopus related errors.

    Attributes
    ----------
    code: :class:`int`
        The error code returned.
    """

    def __init__(self, code):
        self.code = code
        msg = _lib.opus_strerror(self.code).decode('utf-8')
        log.info('"%s" has happened', msg)
        super().__init__(msg)

class OpusNotLoaded(DiscordException):
    """An exception that is thrown for when libopus is not loaded."""
    pass

class _OpusStruct:
    SAMPLING_RATE = 48000
    CHANNELS = 2
    FRAME_LENGTH = 20 # in ms
    SAMPLE_SIZE = 4 # (bit_rate / 8) * CHANNELS (bit_rate == 16)
    SAMPLES_PER_FRAME = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)

    FRAME_SIZE = SAMPLES_PER_FRAME * SAMPLE_SIZE

class Encoder(_OpusStruct):
    def __init__(self, application=APPLICATION_AUDIO):
        if not is_loaded():
            if not _load_default():
                raise OpusNotLoaded()

        self.application = application
        self._state = self._create_state()
        self.set_bitrate(128)
        self.set_fec(True)
        self.set_expected_packet_loss_percent(0.15)
        self.set_bandwidth('full')
        self.set_signal_type('auto')

    def __del__(self):
        if hasattr(self, '_state'):
            _lib.opus_encoder_destroy(self._state)
            self._state = None

    def _create_state(self):
        ret = ctypes.c_int()
        return _lib.opus_encoder_create(self.SAMPLING_RATE, self.CHANNELS, self.application, ctypes.byref(ret))

    def set_bitrate(self, kbps):
        kbps = min(512, max(16, int(kbps)))

        _lib.opus_encoder_ctl(self._state, CTL_SET_BITRATE, kbps * 1024)
        return kbps

    def set_bandwidth(self, req):
        if req not in band_ctl:
            raise KeyError('%r is not a valid bandwidth setting. Try one of: %s' % (req, ','.join(band_ctl)))

        k = band_ctl[req]
        _lib.opus_encoder_ctl(self._state, CTL_SET_BANDWIDTH, k)

    def set_signal_type(self, req):
        if req not in signal_ctl:
            raise KeyError('%r is not a valid signal setting. Try one of: %s' % (req, ','.join(signal_ctl)))

        k = signal_ctl[req]
        _lib.opus_encoder_ctl(self._state, CTL_SET_SIGNAL, k)

    def set_fec(self, enabled=True):
        _lib.opus_encoder_ctl(self._state, CTL_SET_FEC, 1 if enabled else 0)

    def set_expected_packet_loss_percent(self, percentage):
        _lib.opus_encoder_ctl(self._state, CTL_SET_PLP, min(100, max(0, int(percentage * 100))))

    def encode(self, pcm, frame_size):
        max_data_bytes = len(pcm)
        pcm = ctypes.cast(pcm, c_int16_ptr)
        data = (ctypes.c_char * max_data_bytes)()

        ret = _lib.opus_encode(self._state, pcm, frame_size, data, max_data_bytes)

        return array.array('b', data[:ret]).tobytes()

class Decoder(_OpusStruct):
    def __init__(self):
        if not is_loaded():
            raise OpusNotLoaded()

        self._state = self._create_state()

    def __del__(self):
        if hasattr(self, '_state'):
            _lib.opus_decoder_destroy(self._state)
            self._state = None

    def _create_state(self):
        ret = ctypes.c_int()
        return _lib.opus_decoder_create(self.SAMPLING_RATE, self.CHANNELS, ctypes.byref(ret))

    @staticmethod
    def packet_get_nb_frames(data):
        """Gets the number of frames in an Opus packet"""
        return _lib.opus_packet_get_nb_frames(data, len(data))

    @staticmethod
    def packet_get_nb_channels(data):
        """Gets the number of channels in an Opus packet"""
        return _lib.opus_packet_get_nb_channels(data)

    @classmethod
    def packet_get_samples_per_frame(cls, data):
        """Gets the number of samples per frame from an Opus packet"""
        return _lib.opus_packet_get_samples_per_frame(data, cls.SAMPLING_RATE)

    def _set_gain(self, adjustment):
        """Configures decoder gain adjustment.
        Scales the decoded output by a factor specified in Q8 dB units.
        This has a maximum range of -32768 to 32767 inclusive, and returns
        OPUS_BAD_ARG (-1) otherwise. The default is zero indicating no adjustment.
        This setting survives decoder reset (irrelevant for now).

        gain = 10**x/(20.0*256)

        (from opus_defines.h)
        """
        return _lib.opus_decoder_ctl(self._state, CTL_SET_GAIN, adjustment)

    def set_gain(self, dB):
        """Sets the decoder gain in dB, from -128 to 128."""

        dB_Q8 = max(-32768, min(32767, round(dB * 256))) # dB * 2^n where n is 8 (Q8)
        return self._set_gain(dB_Q8)

    def set_volume(self, mult):
        """Sets the output volume as a float percent, i.e. 0.5 for 50%, 1.75 for 175%, etc."""
        return self.set_gain(20 * math.log10(mult)) # amplitude ratio

    def _get_last_packet_duration(self):
        """Gets the duration (in samples) of the last packet successfully decoded or concealed."""

        ret = ctypes.c_int32()
        _lib.opus_decoder_ctl(self._state, CTL_LAST_PACKET_DURATION, ctypes.byref(ret))
        return ret.value

    def decode(self, data, *, fec=False):
        if data is None and fec:
            raise OpusError("Invalid arguments: FEC cannot be used with null data")

        if data is None:
            frame_size = self._get_last_packet_duration() or self.SAMPLES_PER_FRAME
        else:
            frames = self.packet_get_nb_frames(data)
            samples_per_frame = self.packet_get_samples_per_frame(data)
            frame_size = frames * samples_per_frame

        pcm = (ctypes.c_int16 * (frame_size * self.CHANNELS))()
        pcm_ptr = ctypes.cast(pcm, ctypes.POINTER(ctypes.c_int16))

        result = _lib.opus_decode(self._state, data, len(data) if data else 0, pcm_ptr, frame_size, fec)
        return array.array('h', pcm).tobytes()

class BufferedDecoder(threading.Thread):
    DELAY = Decoder.FRAME_LENGTH / 1000.0

    def __init__(self, ssrc, output_func, *, buffer=200):
        super().__init__(daemon=True, name='ssrc-%s' % ssrc)

        if buffer < 40: # technically 20 works but then FEC is useless
            raise ValueError("buffer size of %s is invalid; cannot be lower than 40" % buffer)

        self.ssrc = ssrc
        self.output_func = output_func

        self._decoder = Decoder()
        self._buffer = []
        self._last_seq = 0
        self._last_ts = 0
        self._loops = 0

        # Optional diagnostic state stuff
        self._overflow_mult = self._overflow_base = 2.0
        self._overflow_incr = 0.5

        # minimum (lower bound) size of the jitter buffer (n * 20ms per packet)
        self.buffer_size = buffer // self._decoder.FRAME_LENGTH

        self._finalizing = False
        self._end_thread = threading.Event()
        self._end_main_loop = threading.Event()
        self._primed = threading.Event()
        self._lock = threading.RLock()

        # TODO: Add RTCP queue
        self._rtcp_buffer = []

        self.start()

    def feed_rtp(self, packet):
        if self._last_ts < packet.timestamp:
            self._push(packet)
        elif self._end_thread.is_set():
            return

    def feed_rtcp(self, packet):
        ... # TODO: rotating buffer of Nones or something
        #           or I can store (last_seq + buffer_size, packet)
        # print(f"[router:feed] Got rtcp packet {packet}")
        # print(f"[router:feed] Other timestamps: {[p.timestamp for p in self._buffer]}")
        # print(f"[router:feed] Other timestamps: {self._buffer}")

    def truncate(self, *, size=None):
        """Discards old data to shrink buffer back down to ``size`` (default: buffer_size).
        TODO: doc
        """

        size = self.buffer_size if size is None else size
        with self._lock:
            self._buffer = self._buffer[-size:]

    def stop(self, **kwargs):
        """
        drain=True: continue to write out the remainder of the buffer at the standard rate
        flush=False: write the remainder of the buffer with no delay
        TODO: doc
        """

        with self._lock:
            self._end_thread.set()
            self._end_main_loop.set()

            if any(isinstance(p, RTPPacket) for p in self._buffer) or True:
                if kwargs.pop('flush', False):
                    self._finalizing = True
                    self.DELAY = 0
                elif not kwargs.pop('drain', True):
                    with self._lock:
                        self._finalizing = True
                        self._buffer.clear()

    def reset(self):
        with self._lock:
            self._decoder = Decoder() # TODO: Add a reset function to Decoder itself
            self._last_seq = self._last_ts = 0
            self._buffer.clear()
            self._primed.clear()
            self._end_main_loop.set() # XXX: racy with _push?
            self.DELAY = self.__class__.DELAY

    def _push(self, item):
        if not isinstance(item, (RTPPacket, SilencePacket)):
            raise TypeError(f"item should be an RTPPacket, not {item.__class__.__name__}")

        # XXX: racy with reset?
        if self._end_main_loop.is_set() and not self._end_thread.is_set():
            self._end_main_loop.clear()

        if not self._primed.is_set():
            self._primed.set()

        # Fake packet loss
        # import random
        # if random.randint(1, 100) <= 10 and isinstance(item, RTPPacket):
        #     return

        with self._lock:
            existing_packet = utils.get(self._buffer, timestamp=item.timestamp)
            if isinstance(existing_packet, SilencePacket):
                # Replace silence packets with rtp packets
                self._buffer[self._buffer.index(existing_packet)] = item
                return
            elif isinstance(existing_packet, RTPPacket):
                return # duplicate packet

            bisect.insort(self._buffer, item)

        # Optional diagnostics, will probably remove later
            bufsize = len(self._buffer) # indent intentional
        if bufsize >= self.buffer_size * self._overflow_mult:
            print(f"[router:push] Warning: rtp heap size has grown to {bufsize}")
            self._overflow_mult += self._overflow_incr

        elif bufsize <= self.buffer_size * (self._overflow_mult - self._overflow_incr) \
            and self._overflow_mult > self._overflow_base:

            print(f"[router:push] Info: rtp heap size has shrunk to {bufsize}")
            self._overflow_mult = max(self._overflow_base, self._overflow_mult - self._overflow_incr)

    def _pop(self):
        packet = nextpacket = None
        with self._lock:
            try:
                if not self._finalizing:
                    self._buffer.append(SilencePacket(self.ssrc, self._buffer[-1].timestamp + Decoder.SAMPLES_PER_FRAME))
                packet = self._buffer.pop(0)
                nextpacket = self._buffer[0]
            except IndexError:
                pass # empty buffer

        return packet, nextpacket

    def _initial_fill(self):
        """Artisanal hand-crafted function for buffering packets and clearing discord's stupid fucking rtp buffer."""

        if self._end_main_loop.is_set():
            return

        # Very small sleep to check if there's buffered packets
        time.sleep(0.001)
        if len(self._buffer) > 3:
            # looks like there's some old packets in the buffer
            # we need to figure out where the old packets stop and where the fresh ones begin
            # for that we need to see when we return to the normal packet accumulation rate

            last_size = len(self._buffer)

            # wait until we have the correct rate of packet ingress
            while len(self._buffer) - last_size > 1:
                last_size = len(self._buffer)
                time.sleep(0.001)

            # collect some fresh packets
            time.sleep(0.06)

            # generate list of differences between packet sequences
            with self._lock:
                diffs = [self._buffer[i+1].sequence-self._buffer[i].sequence for i in range(len(self._buffer)-1)]
            sdiffs = sorted(diffs, reverse=True)

            # decide if there's a jump
            jump1, jump2 = sdiffs[:2]
            if jump1 > jump2 * 3:
                # remove the stale packets and keep the fresh ones
                self.truncate(size=len(self._buffer[diffs.index(jump1)+1:]))
            else:
                # otherwise they're all stale, dump 'em (does this ever happen?)
                with self._lock:
                    self._buffer.clear()

        # fill buffer to at least half full
        while len(self._buffer) < self.buffer_size // 2:
            time.sleep(0.001)

        # fill the buffer with silence aligned with the first packet
        # if an rtp packet already exists for the given silence packet ts, the silence packet is ignored
        with self._lock:
            start_ts = self._buffer[0].timestamp
            for x in range(1, 1 + self.buffer_size - len(self._buffer)):
                self._push(SilencePacket(self.ssrc, start_ts + x * Decoder.SAMPLES_PER_FRAME))

        # now fill the rest
        while len(self._buffer) < self.buffer_size:
            time.sleep(0.001)
            # TODO: Maybe only wait at most for about as long we we're supposed to?
            #       0.02 * (buffersize - len(buffer))

    def _packet_gen(self):
        while True:
            packet, nextpacket = self._pop()
            self._last_ts = getattr(packet, 'timestamp', self._last_ts + Decoder.SAMPLES_PER_FRAME)
            self._last_seq += 1 # self._last_seq = packet.sequence?

            if isinstance(packet, RTPPacket):
                pcm = self._decoder.decode(packet.decrypted_data)

            elif isinstance(nextpacket, RTPPacket):
                pcm = self._decoder.decode(packet.decrypted_data, fec=True)
                fec_packet = FECPacket(self.ssrc, nextpacket.sequence - 1, nextpacket.timestamp - Decoder.SAMPLES_PER_FRAME)
                yield fec_packet, pcm

                packet, _ = self._pop()
                self._last_ts += Decoder.SAMPLES_PER_FRAME
                self._last_seq += 1

                pcm = self._decoder.decode(packet.decrypted_data)

            elif packet is None:
                self._finalizing = False
                break
            else:
                pcm = self._decoder.decode(None)

            yield packet, pcm

    def _do_run(self):
        self._primed.wait()
        self._initial_fill()

        self._loops = 0
        packet_gen = self._packet_gen()
        start_time = time.perf_counter()
        try:
            while not self._end_main_loop.is_set() or self._finalizing:
                packet, pcm = next(packet_gen)
                try:
                    self.output_func(pcm, packet.decrypted_data, packet)
                except:
                    log.exception("Sink raised exception")
                    traceback.print_exc()

                next_time = start_time + self.DELAY * self._loops
                self._loops += 1

                time.sleep(max(0, self.DELAY + (next_time - time.perf_counter())))
        except StopIteration:
            time.sleep(0.001) # just in case, so we don't slam the cpu
        finally:
            packet_gen.close()

    def run(self):
        try:
            while not self._end_thread.is_set():
                self._do_run()
        except Exception as e:
            log.exception("Error in decoder %s", self.name)
            traceback.print_exc()


class BasePacketDecoder:
    DELAY = Decoder.FRAME_LENGTH / 1000.0

    def feed_rtp(self, packet):
        raise NotImplementedError
    def feed_rtcp(self, packet):
        raise NotImplementedError
    def truncate(self, *, size=None):
        raise NotImplementedError
    def reset(self):
        raise NotImplementedError

class BufferedPacketDecoder(BasePacketDecoder):
    """Buffers and decodes packets from a single ssrc"""

    def __init__(self, ssrc, *, buffer=200):
        if buffer < 40: # technically 20 works but then FEC is useless
            raise ValueError("buffer size of %s is invalid; cannot be lower than 40" % buffer)

        self.ssrc = ssrc
        self._decoder = Decoder()
        self._buffer = []
        self._rtcp_buffer = {} # TODO: Add RTCP queue
        self._last_seq = self._last_ts = 0

        # Optional diagnostic state stuff
        self._overflow_mult = self._overflow_base = 2.0
        self._overflow_incr = 0.5

        # minimum (lower bound) size of the jitter buffer (n * 20ms per packet)
        self.buffer_size = buffer // self._decoder.FRAME_LENGTH
        self._lock = threading.RLock()

        self._gen = None

    def __iter__(self):
        if self._gen is None:
            self._gen = self._packet_gen()
        return self._gen

    def __next__(self):
        return next(iter(self))

    def feed_rtp(self, packet):
        if self._last_ts < packet.timestamp:
            self._push(packet)

    def feed_rtcp(self, packet):
        with self._lock:
            if not self._buffer:
                return # ignore for now, handle properly later
            self._rtcp_buffer[self._buffer[-1]] = packet

    def truncate(self, *, size=None):
        size = self.buffer_size if size is None else size
        with self._lock:
            self._buffer = self._buffer[-size:]

    def reset(self):
        with self._lock:
            self._decoder = Decoder() # TODO: Add a reset function to Decoder itself
            self.DELAY = self.__class__.DELAY
            self._last_seq = self._last_ts = 0
            self._buffer.clear()
            self._rtcp_buffer.clear()
            self._gen.close()
            self._gen = None

    def _push(self, item):
        if not isinstance(item, (RTPPacket, SilencePacket)):
            raise TypeError(f"item should be an RTPPacket, not {item.__class__.__name__}")

        # Fake packet loss
        # import random
        # if random.randint(1, 100) <= 10 and isinstance(item, RTPPacket):
        #     return

        with self._lock:
            existing_packet = utils.get(self._buffer, timestamp=item.timestamp)
            if isinstance(existing_packet, SilencePacket):
                # Replace silence packets with rtp packets
                self._buffer[self._buffer.index(existing_packet)] = item
                return
            elif isinstance(existing_packet, RTPPacket):
                return # duplicate packet

            bisect.insort(self._buffer, item)

        # Optional diagnostics, will probably remove later
            bufsize = len(self._buffer) # indent intentional
        if bufsize >= self.buffer_size * self._overflow_mult:
            print(f"[router:push] Warning: rtp heap size has grown to {bufsize}")
            self._overflow_mult += self._overflow_incr

        elif bufsize <= self.buffer_size * (self._overflow_mult - self._overflow_incr) \
            and self._overflow_mult > self._overflow_base:

            print(f"[router:push] Info: rtp heap size has shrunk to {bufsize}")
            self._overflow_mult = max(self._overflow_base, self._overflow_mult - self._overflow_incr)

    def _pop(self):
        packet = nextpacket = None
        with self._lock:
            try:
                self._buffer.append(SilencePacket(self.ssrc, self._buffer[-1].timestamp + Decoder.SAMPLES_PER_FRAME))
                packet = self._buffer.pop(0)
                nextpacket = self._buffer[0]
            except IndexError:
                pass # empty buffer

        return packet, nextpacket # return rtcp packets as well?

    def _packet_gen(self):
        # Buffer packets
        # do I care about dumping buffered packets on reset?

        # Ok yes this section is going to look weird.  To keep everything consistant I need to
        # wait for a specific number of iterations instead of on the actual buffer size.  These
        # objects are supposed to be time naive.  The class handling these is responsible for
        # keeping the time synchronization.

        # How many packets we already have
        pre_fill = len(self._buffer)
        # How many packets we need to get to half full
        half_fill = max(0, self.buffer_size//2 - 1 - pre_fill)
        # How many packets we need to get to full
        full_fill = self.buffer_size - half_fill

        print(f"Starting with {pre_fill}, collecting {half_fill}, then {full_fill}")

        while not self._buffer:
            yield None, None

        for x in range(half_fill-1):
            yield None, None

        with self._lock:
            start_ts = self._buffer[0].timestamp
            for x in range(1, 1 + self.buffer_size - len(self._buffer)):
                self._push(SilencePacket(self.ssrc, start_ts + x * Decoder.SAMPLES_PER_FRAME))

        for x in range(full_fill):
            yield None, None

        while True:
            packet, nextpacket = self._pop()
            self._last_ts = getattr(packet, 'timestamp', self._last_ts + Decoder.SAMPLES_PER_FRAME)
            self._last_seq += 1 # self._last_seq = packet.sequence?

            if isinstance(packet, RTPPacket):
                pcm = self._decoder.decode(packet.decrypted_data)

            elif isinstance(nextpacket, RTPPacket):
                pcm = self._decoder.decode(packet.decrypted_data, fec=True)
                fec_packet = FECPacket(self.ssrc, nextpacket.sequence - 1, nextpacket.timestamp - Decoder.SAMPLES_PER_FRAME)
                yield fec_packet, pcm

                packet, _ = self._pop()
                self._last_ts += Decoder.SAMPLES_PER_FRAME
                self._last_seq += 1

                pcm = self._decoder.decode(packet.decrypted_data)

            elif packet is None:
                break
            else:
                pcm = self._decoder.decode(None)

            yield packet, pcm


class BufferedDecoder(threading.Thread):
    """Ingests rtp packets and dispatches to decoders and sink output function."""

    def __init__(self, reader, *, decodercls=BufferedPacketDecoder):
        super().__init__(daemon=True, name='DecoderBuffer')
        self.reader = reader
        self.decodercls = decodercls

        self.output_func = reader._write_to_sink
        self.decoders = {}
        self.initial_buffer = []
        self.queue = deque()

        self._end_thread = threading.Event()
        self._has_decoder = threading.Event()
        self._lock = threading.Lock()

    def _get_decoder(self, ssrc):
        dec = self.decoders.get(ssrc)

        if not dec and self.reader.client._get_ssrc_mapping(ssrc=ssrc)[1]: # and get_user(ssrc)?
            dec = self.decoders[ssrc] = self.decodercls(ssrc)
            dec.start_time = time.perf_counter() # :thinking:
            dec.loops = 0                        # :thinking::thinking::thinking:
            self.queue.append((dec.start_time, dec))
            self._has_decoder.set()

        return dec

    def _feed_rtp_initial(self, packet):
        with self._lock:
            self.initial_buffer.append(packet)

    def feed_rtp(self, packet):
        dec = self._get_decoder(packet.ssrc)
        if dec:
            return dec.feed_rtp(packet)

    def feed_rtcp(self, packet):
        # RTCP packets themselves don't really belong to a decoder
        # I could split the reports up or send to all idk its weird

        dec = self._get_decoder(packet.ssrc)
        if dec:
            print(f"RTCP packet: {packet}")
            return dec.feed_rtcp(packet)

    def drop_ssrc(self, ssrc):
        dec = self.decoders.pop(ssrc, None)
        if dec:
            # dec/self.flush()?
            dec.reset()

            if not self.decoders:
                self._has_decoder.clear()

    def reset(self, *ssrcs):
        with self._lock:
            if not ssrcs:
                ssrcs = tuple(self.decoders.keys())

            for ssrc in ssrcs:
                dec = self.decoders.get(ssrc)
                if dec:
                    dec.reset()

    def flush(self, *ssrcs):
        ...
        # The new idea is to call a special flush event function on the sink with the
        # rest of the audio buffer when exiting so the user can use or ignore it

    def stop(self, **kwargs):
        for decoder in tuple(self.decoders.values()):
            # decoder.stop(**kwargs)
            decoder.reset()

    def _initial_fill(self):
        # Fill a single buffer first then dispense into the actual buffers
        try:
            normal_feed_rtp = self.feed_rtp
            self.feed_rtp = self._feed_rtp_initial

            buff = self.initial_buffer

            # Very small sleep to check if there's buffered packets
            time.sleep(0.002)
            if len(buff) > 3:
                # looks like there's some old packets in the buffer
                # we need to figure out where the old packets stop and where the fresh ones begin
                # for that we need to see when we return to the normal packet accumulation rate

                last_size = len(buff)

                # wait until we have the correct rate of packet ingress
                while len(buff) - last_size > 1:
                    last_size = len(buff)
                    time.sleep(0.001)

                # collect some fresh packets
                time.sleep(0.06)

                # generate list of differences between packet sequences
                with self._lock:
                    diffs = [buff[i+1].sequence - buff[i].sequence for i in range(len(buff)-1)]
                sdiffs = sorted(diffs, reverse=True)

                # decide if there's a jump
                jump1, jump2 = sdiffs[:2]
                if jump1 > jump2 * 3:
                    # remove the stale packets and keep the fresh ones
                    with self._lock:
                        size = len(buff[diffs.index(jump1)+1:])
                        buff = buff[-size:]
                else:
                    # otherwise they're all stale, dump 'em (does this ever happen?)
                    with self._lock:
                        buff.clear()

            # The old version of this code backfilled buffers based on the buffer size.
            # We dont have that here but we can just have the individual buffer objects
            # backfill themselves.

            # Dump initial buffer into actual buffers
            with self._lock:
                for packet in buff:
                    normal_feed_rtp(packet)

                self.feed_rtp = normal_feed_rtp
        finally:
            self.feed_rtp = normal_feed_rtp

    def decode(self, decoder):
        data = next(decoder)
        if any(data):
            packet, pcm = data
            try:
                self.output_func(pcm, packet.decrypted_data, packet)
            except:
                log.exception("Sink raised exception")
                traceback.print_exc()

        decoder.loops += 1
        decoder.next_time = decoder.start_time + decoder.DELAY * decoder.loops
        self.queue.append((decoder.next_time, decoder))

    def _do_run(self):
        while not self._end_thread.is_set():
            self._has_decoder.wait()

            next_time, decoder = self.queue.popleft()
            remaining = next_time - time.perf_counter()

            if remaining >= 0:
                insort(self.queue, (next_time, decoder))
                time.sleep(max(0.002, remaining/2)) # sleep accuracy tm
                continue

            self.decode(decoder)

    def run(self):
        try:
            self._do_run()
        except Exception as e:
            log.exception("Error in decoder %s", self.name)
            traceback.print_exc()



from flask import Blueprint, render_template, request, Response, jsonify, redirect
from typing import List

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.web_interface.auth.models import AccountData

auth_blueprint = Blueprint("auth", __name__, template_folder="templates")

@auth_blueprint.route('/login')
def login():
    engine_resources = EngineResources()
    account_resources = AccountResources(engine_resources=engine_resources)

    login_success = account_resources.login_with_cookies(cookies=request.cookies)
    if login_success is not True:
        return render_template("auth/login.html")
    else:
        return redirect(f"{account_resources.account_username}/control-panel")

@auth_blueprint.route('/auth_backend/login', methods=["POST"])
def auth_backend_login():
    engine_resources = EngineResources()
    account_resources = AccountResources(engine_resources=engine_resources)

    data = SafeDict(request.get_json())
    username_or_email = data.get("usernameOrEmail").to_str(default=None)
    encrypted_password = data.get("encryptedPassword").to_str(default=None)

    if username_or_email is not None and encrypted_password is not None:
        account_has_been_found, account_has_been_logged_in, new_auth_token_value = account_resources.login(
            username_or_email=username_or_email, encrypted_password=encrypted_password
        )

        if account_has_been_logged_in is True:
            print(new_auth_token_value)
            return jsonify({
                "loginSuccess": True,
                "accountUsername": account_resources.account_username,
                "accountId": account_resources.account_id,
                "newAuthToken": new_auth_token_value
            })
        else:
            return jsonify({
                "loginSuccess": False
            })
    else:
        # todo: change that to have the same format everywhere
        if username_or_email is None and encrypted_password is None:
            return Response("mailAndPasswordFieldsEmpty", status=401)
        elif username_or_email is None:
            return Response("mailFieldEmpty", status=401)
        elif username_or_email is None:
            return Response("passwordFieldEmpty", status=401)

@auth_blueprint.route('/auth_backend/login-with-amazon', methods=["POST"])
def login_with_amazon():
    authorization_code = SafeDict(request.get_json()).get("authorizationCode").to_str(default=None)
    if authorization_code is not None:
        from inoft_vocal_engine.web_interface.auth.login_with_amazon import LoginWithAmazon
        tokens = LoginWithAmazon().authorization_code_to_tokens(authorization_code=authorization_code)

        from inoft_vocal_engine.smapi_clients.smapi_base_client import SmapiBaseClient
        smapi_client = SmapiBaseClient(refresh_token=tokens.refresh_token)
        skills_summaries = smapi_client.get_list_skills_of_account()

        skills_dicts: List[dict] = list()
        for skill in skills_summaries:
            skills_dicts.append(skill.to_dict())

        print(skills_dicts)

        smapi_client.create_skill(vendor_id=smapi_client.get_vendors_of_account()[0].id)

        return jsonify(success=True, skills=skills_dicts)
    return jsonify(success=False)





from typing import List, Optional

from pydantic import BaseModel
from pydantic.dataclasses import dataclass


class FlowFileBotpressModel(BaseModel):
    class NodeModel(BaseModel):
        id: str
        name: str = None
        type: str = None
        next: Optional[List[dict]] = None
        onEnter: Optional[List[str]] = None
        onReceive: Optional[List[str]] = None
    version: str
    catchAll: dict
    startNode: str
    description: Optional[str] = None
    nodes: Optional[List[NodeModel]] = list()

class UiFileBotpressModel(BaseModel):
    class NodePositionModel(BaseModel):
        class PositionModel(BaseModel):
            x: int
            y: int
        id: str
        position: Optional[PositionModel] = PositionModel(x=0, y=0)
    nodes: Optional[List[NodePositionModel]] = None



from jinja2 import BaseLoader, TemplateNotFound
from os.path import exists, getmtime

class JinjaFilepathLoader(BaseLoader):
    def __init__(self, filepath):
        self.filepath = filepath

    def get_source(self, environment, template):
        if not exists(self.filepath):
            raise TemplateNotFound(template)
        mtime = getmtime(self.filepath)
        with open(self.filepath) as f:
            source = f.read()
        return source, self.filepath, lambda: mtime == getmtime(self.filepath)



import boto3
from boto3.dynamodb.conditions import Key, Attr
from boto3.exceptions import ResourceNotExistsError
from boto3.session import Session
from typing import List, Optional, Type, Any, Dict

from botocore.exceptions import ClientError
from pydantic import BaseModel, validate_arguments
from pydantic.dataclasses import dataclass

from inoft_vocal_engine.databases.dynamodb.dynamodb_utils import Utils
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.utils.static_logger import logger

HASH_KEY_TYPE = "HASH"
SORT_KEY_TYPE = "RANGE"


class GetItemResponse(BaseModel):
    item: Optional[dict]
    success: bool

class Response:
    def __init__(self, response_dict: dict):
        response_safedict = SafeDict(response_dict)
        self.items = response_safedict.get("Items").to_list(default=None)
        self.count = response_safedict.get("Count").to_int(default=None)
        self.scanned_count = response_safedict.get("ScannedCount").to_int(default=None)
        self.last_evaluated_key = response_safedict.get("LastEvaluatedKey").to_dict(default=None)
        self.has_reached_end = False if self.last_evaluated_key is not None else True

class Index(BaseModel):
    hash_key_name: str
    hash_key_variable_python_type: Type
    sort_key_name: Optional[str]
    sort_key_variable_python_type: Optional[Type]
    index_custom_name: Optional[str]


class PrimaryIndex(Index):
    pass

class GlobalSecondaryIndex(Index):
    # todo: add pydantic to this class
    PROJECTION_TYPE_USE_ALL = "ALL"
    PROJECTION_TYPE_KEYS_ONLY = "KEYS_ONLY"
    PROJECTION_TYPE_INCLUDE = "INCLUDE"
    ALL_PROJECTIONS_TYPES = [PROJECTION_TYPE_USE_ALL, PROJECTION_TYPE_KEYS_ONLY, PROJECTION_TYPE_INCLUDE]

    projection_type: str

    def __init__(self, projection_type: str, hash_key_name: str, hash_key_variable_python_type: Type,
                 sort_key_name: Optional[str] = None, sort_key_variable_python_type: Optional[Type] = None,
                 index_custom_name: Optional[str] = None):

        super().__init__(hash_key_name=hash_key_name, hash_key_variable_python_type=hash_key_variable_python_type,
                         sort_key_name=sort_key_name, sort_key_variable_python_type=sort_key_variable_python_type,
                         projection_type=projection_type)

        if projection_type not in self.ALL_PROJECTIONS_TYPES:
            raise Exception(f"{projection_type} has not been found in the available projection_types : {self.ALL_PROJECTIONS_TYPES}")
        self.projection_type = projection_type

    def to_dict(self):
        if self.index_custom_name is not None:
            index_name = self.index_custom_name
        else:
            if self.sort_key_name is None:
                index_name = self.hash_key_name
            else:
                index_name = f"{self.hash_key_name}-{self.sort_key_name}"

        output_dict = {
            "IndexName": index_name,
            "KeySchema": [
                {
                    "AttributeName": self.hash_key_name,
                    "KeyType": HASH_KEY_TYPE
                },
            ],
            "Projection": {
                "ProjectionType": self.projection_type
            }
        }
        if self.sort_key_name is not None and self.sort_key_variable_python_type is not None:
            output_dict["KeySchema"].append({
                "AttributeName": self.sort_key_name,
                "KeyType": SORT_KEY_TYPE
            })
        return output_dict


class CreateTableQueryKwargs:
    @validate_arguments
    def __init__(self, table_name: str, billing_mode: str):
        self._names_already_defined_attributes: List[str] = list()
        self.data = {
            "TableName": table_name,
            "KeySchema": list(),
            "AttributeDefinitions": list(),
            "BillingMode": billing_mode,
        }

    def _add_key(self, key_name: str, key_python_variable_type: Type, key_type: str):
        self.data["KeySchema"].append({
            "AttributeName": key_name,
            "KeyType": key_type
        })
        self.data["AttributeDefinitions"].append({
            "AttributeName": key_name,
            "AttributeType": Utils.python_type_to_dynamodb_type(python_type=key_python_variable_type)
        })

    @validate_arguments
    def add_hash_key(self, key_name: str, key_python_variable_type: Type):
        self._add_key(key_name=key_name, key_python_variable_type=key_python_variable_type, key_type=HASH_KEY_TYPE)

    @validate_arguments
    def add_sort_key(self, key_name: str, key_python_variable_type: Type):
        self._add_key(key_name=key_name, key_python_variable_type=key_python_variable_type, key_type=SORT_KEY_TYPE)

    def _add_global_secondary_index(self, key_name: str, key_python_variable_type: Type):
        if key_name not in self._names_already_defined_attributes:
            self.data["AttributeDefinitions"].append({
                "AttributeName": key_name,
                "AttributeType": Utils.python_type_to_dynamodb_type(python_type=key_python_variable_type)
            })
            self._names_already_defined_attributes.append(key_name)

    @validate_arguments
    def add_global_secondary_index(self, global_secondary_index: GlobalSecondaryIndex):
        if "GlobalSecondaryIndexes" not in self.data:
            self.data["GlobalSecondaryIndexes"] = list()

        self._add_global_secondary_index(key_name=global_secondary_index.hash_key_name,
                                         key_python_variable_type=global_secondary_index.hash_key_variable_python_type)
        if global_secondary_index.sort_key_name is not None and global_secondary_index.sort_key_variable_python_type is not None:
            self._add_global_secondary_index(key_name=global_secondary_index.sort_key_name,
                                             key_python_variable_type=global_secondary_index.sort_key_variable_python_type)

        self.data["GlobalSecondaryIndexes"].append(global_secondary_index.to_dict())

    @validate_arguments
    def add_all_global_secondary_indexes(self, global_secondary_indexes: List[GlobalSecondaryIndex]):
        for global_secondary_index in global_secondary_indexes:
            self.add_global_secondary_index(global_secondary_index=global_secondary_index)


@dataclass
class DynamoDBMapObjectSetter:
    object_path: str
    value_to_set: Any
    map_key: Optional[str] = None


class DynamoDbCoreAdapter:
    _EXISTING_DATABASE_CLIENTS = dict()
    PAY_PER_REQUEST = "PAY_PER_REQUEST"
    PROVISIONED = "PROVISIONED"

    def __init__(self, table_name: str, region_name: str, primary_index: PrimaryIndex,
                 create_table: bool = True, billing_mode: str = PAY_PER_REQUEST,
                 global_secondary_indexes: List[GlobalSecondaryIndex] = None):
        self.table_name = table_name
        self.primary_index = primary_index
        self.create_table = create_table
        self.billing_mode = billing_mode
        self.global_secondary_indexes = global_secondary_indexes

        # We store the database clients in a static variable, so that if we init the class with
        # the same region_name, we do not need to wait for a new initialization of the client.
        if region_name in self._EXISTING_DATABASE_CLIENTS.keys():
            self.dynamodb = self._EXISTING_DATABASE_CLIENTS[region_name]
            print(f"Re-using the already created dynamodb client for region {region_name}")
        elif "default" in self._EXISTING_DATABASE_CLIENTS.keys():
            self.dynamodb = self._EXISTING_DATABASE_CLIENTS["default"]
            print(f"Re-using the already created dynamodb client for the default region")
        else:
            print(f"Initializing the {self}. For local development, make sure that you are connected to internet."
                  f"\nOtherwise the framework will get stuck at initializing the {self}")

            dynamodb_regions = Session().get_available_regions("dynamodb")
            if region_name in dynamodb_regions:
                self.dynamodb = boto3.resource("dynamodb", region_name=region_name)
                self._EXISTING_DATABASE_CLIENTS[region_name] = self.dynamodb
            else:
                self.dynamodb = boto3.resource("dynamodb")
                self._EXISTING_DATABASE_CLIENTS["default"] = self.dynamodb
                logger.debug(f"Warning ! The specified dynamodb region_name {region_name} is not a valid region_name."
                             f"The dynamodb client has been initialized without specifying the region.")

        self._create_table_if_not_exists()
        print(f"Initialization of {self} completed successfully !")

    def _create_table_if_not_exists(self) -> None:
        """
        Creates table in Dynamodb resource if it doesn't exist and create_table is set as True.
        :raises: PersistenceException: When `create_table` fails on dynamodb resource.
        """
        if self.create_table:
            create_table_query_kwargs = CreateTableQueryKwargs(table_name=self.table_name, billing_mode=self.billing_mode)

            create_table_query_kwargs.add_hash_key(key_name=self.primary_index.hash_key_name,
                                                   key_python_variable_type=self.primary_index.hash_key_variable_python_type)

            if self.primary_index.sort_key_name is not None and self.primary_index.sort_key_variable_python_type is not None:
                create_table_query_kwargs.add_hash_key(key_name=self.primary_index.sort_key_name,
                                                       key_python_variable_type=self.primary_index.sort_key_variable_python_type)

            if self.global_secondary_indexes is not None:
                create_table_query_kwargs.add_all_global_secondary_indexes(global_secondary_indexes=self.global_secondary_indexes)
            try:
                print(create_table_query_kwargs.data)
                self.dynamodb.create_table(**create_table_query_kwargs.data)
            except Exception as e:
                if e.__class__.__name__ != "ResourceInUseException":
                    raise Exception(f"Create table if not exists request failed: Exception of type {type(e).__name__} occurred {str(e)}")

    def _put_item_dict(self, item_dict: dict) -> Response:
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.put_item(Item=item_dict)
            return response
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save attributes to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def _get_item_by_primary_key(self, key_name: str, key_value: any, fields_to_get: Optional[List[str]]) -> Optional[GetItemResponse]:
        if fields_to_get is not None:
            kwargs = self._fields_to_get_to_expressions(fields_to_get=fields_to_get)
        else:
            kwargs = dict()
        kwargs["Key"] = {key_name: key_value}
        kwargs["ConsistentRead"] = True

        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.get_item(**kwargs)
            if "Item" in response:
                return GetItemResponse(item=Utils.dynamodb_to_python(dynamodb_object=response["Item"]), success=True)
            else:
                return GetItemResponse(item=None, success=False)
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")

    def _execute_update_query(self, query_kwargs_dict: dict) -> Optional[Response]:
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.update_item(**query_kwargs_dict)
            return Response(response)
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except ClientError as e:
            print(f"{e} - No element has been found for the update query : {query_kwargs_dict}")
            return None
        except Exception as e:
            print(type(e))
            raise Exception(f"Failed to update attributes in DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")

    def _add_data_elements_to_list(self, key_name: str, key_value: Any, object_path: str,
                                   element_values: List[dict]) -> Optional[Response]:
        kwargs = {
            "TableName": self.table_name,
            "Key": {key_name: key_value},
            "ReturnValues": "UPDATED_NEW",
            "UpdateExpression": f"SET {object_path} = list_append(if_not_exists({object_path}, :emptyList), :newItems)",
            # The if_not_exists inside the list_append, will create an empty
            # list before adding the newItems, only if the field do not exist.
            "ExpressionAttributeValues": {
                ":newItems": element_values,
                ":emptyList": []
            }
        }
        return self._execute_update_query(query_kwargs_dict=kwargs)

    def _remove_data_elements_from_list(self, key_name: str, key_value: Any, list_object_path: str,
                                        indexes_to_remove: list) -> Optional[Response]:
        kwargs = {
            "TableName": self.table_name,
            "Key": {key_name: key_value},
            "ReturnValues": "UPDATED_NEW"
        }
        update_expression = "REMOVE "
        for i, index_in_database_list in enumerate(indexes_to_remove):
            update_expression += f"{list_object_path}[{index_in_database_list}]"
            if i + 1 < len(indexes_to_remove):
                update_expression += ", "
        kwargs["UpdateExpression"] = update_expression

        return self._execute_update_query(query_kwargs_dict=kwargs)

    def _initialize_all_elements_in_map_target(self, key_name: str, key_value: Any, object_path_elements: Dict[str, type]) -> bool:
        current_path_target = ""
        expression_attribute_names_dict: Dict[str, str] = dict()
        for i, path_element_key in enumerate(object_path_elements):
            if i > 0:
                current_path_target += "."

            current_path_key = f"#pathKey{i}"
            current_path_target += current_path_key
            expression_attribute_names_dict[current_path_key] = path_element_key

            current_update_expression = f"SET {current_path_target} = if_not_exists({current_path_target}, :item)"
            current_path_initialization_variable_type: type = object_path_elements[path_element_key]
            current_set_potentially_missing_object_query_kwargs = {
                "TableName": self.table_name,
                "Key": {key_name: key_value},
                "ReturnValues": "UPDATED_NEW",
                "UpdateExpression": current_update_expression,
                "ExpressionAttributeNames": expression_attribute_names_dict,
                "ExpressionAttributeValues": {
                    ":item": current_path_initialization_variable_type()
                }
            }
            response = self._execute_update_query(query_kwargs_dict=current_set_potentially_missing_object_query_kwargs)
            if response is None:
                return False
        return True

    def _add_update_data_element_to_map(self, key_name: str, key_value: Any,
                                        object_path_elements: Dict[str, type], element_values: Any) -> Optional[Response]:
        expression_attribute_names_dict = dict()
        update_expression = "SET "

        for i, path_element in enumerate(object_path_elements):
            current_path_key = f"#pathKey{i}"
            update_expression += current_path_key
            expression_attribute_names_dict[current_path_key] = path_element
            if i + 1 < len(object_path_elements):
                update_expression += "."
            else:
                update_expression += " = :item"

        update_query_kwargs = {
            "TableName": self.table_name,
            "Key": {key_name: key_value},
            "ReturnValues": "UPDATED_NEW",
            "UpdateExpression": update_expression,
            "ExpressionAttributeNames": expression_attribute_names_dict,
            "ExpressionAttributeValues": {
                ":item": element_values
            }
        }
        response = self._execute_update_query(query_kwargs_dict=update_query_kwargs)
        if response is None:
            # If the response is None, it means that one of the path of the
            # target path has not been found and need to be initialized.
            success: bool = self._initialize_all_elements_in_map_target(
                key_name=key_name, key_value=key_value, object_path_elements=object_path_elements
            )
            if success is True:
                response = self._execute_update_query(query_kwargs_dict=update_query_kwargs)
        return response

    def _add_update_multiple_data_elements_to_map(self, key_name: str, key_value: Any,
                                                  objects_setters: List[DynamoDBMapObjectSetter]) -> Optional[Response]:
        if not len(objects_setters) > 0:
            # If no object setter has been defined,
            # the query will crash when executed.
            return Response({})

        kwargs = {
            "TableName": self.table_name,
            "Key": {key_name: key_value},
            "ReturnValues": "UPDATED_NEW"
        }

        update_expression = "SET "
        attribute_names_expression_dict = dict()
        attribute_values_expression_dict = dict()
        for i, current_object in enumerate(objects_setters):
            # We use the attribute_names and attribute_values expressions instead of putting the data
            # directly in the update_expression, because boto3 will only convert the data in the
            # appropriate format when it is put in the attribute and values expression.
            # It will not convert it if the data is directly set in the update expression.
            update_expression += f"{current_object.object_path}"

            if current_object.map_key is not None:
                update_expression += f".#key{i}"
                attribute_names_expression_dict[f"#key{i}"] = current_object.map_key

            update_expression += f" = :item{i}"
            attribute_values_expression_dict[f":item{i}"] = current_object.value_to_set

            if i + 1 < len(objects_setters):
                update_expression += ", "

        kwargs["UpdateExpression"] = update_expression
        kwargs["ExpressionAttributeValues"] = attribute_values_expression_dict
        if len(attribute_names_expression_dict) > 0:
            kwargs["ExpressionAttributeNames"] = attribute_names_expression_dict

        return self._execute_update_query(query_kwargs_dict=kwargs)

    def _query_by_key(self, key_name: str, key_value: Any, index_name: Optional[str] = None,
                      fields_to_get: Optional[list] = None, filter_expression: Optional[Any] = None,
                      query_limit: Optional[int] = None) -> Response:

        if fields_to_get is not None:
            kwargs = self._fields_to_get_to_expressions(fields_to_get=fields_to_get)
        else:
            kwargs = dict()

        kwargs["KeyConditionExpression"] = Key(key_name).eq(key_value)
        if index_name is not None:
            kwargs["IndexName"] = index_name

        if filter_expression is not None:
            kwargs["FilterExpression"] = filter_expression

        if query_limit is not None:
            kwargs["Limit"] = query_limit

        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.query(**kwargs)
            return Response(Utils.dynamodb_to_python(response))
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")

    def _query_single_item_by_key(self, key_name: str, key_value: Any, index_name: Optional[str] = None,
                                  fields_to_get: Optional[list] = None, filter_expression: Optional[Any] = None) -> Optional[dict]:
        # Yes, a query request is heavier than a get request that we could do with the _get_item_by_primary_key function.
        # Yet, in a get request, we cannot specify an index_name to query on. So, the _query_single_item_by_key should be
        # used when we want to get an item based on another index that the primary one. Otherwise, use _get_item_by_primary_key
        response = self._query_by_key(
            key_name=key_name, key_value=key_value, index_name=index_name,
            fields_to_get=fields_to_get, filter_expression=filter_expression, query_limit=1
        )
        if response.count == 1:
            return response.items[0]
        elif not response.count > 0:
            print("No item has been found")
            return None
        elif not response.count > 1:
            print("More than one item has been found. Returning first item.")
            return response.items[0]

    def _get_data_in_path_target(self, key_name: str, key_value: str, target_path_elements: Dict[str, type],
                                 num_keys_to_navigation_into: int) -> Optional[any]:
        target_field = ""
        for i, key_path_element in enumerate(target_path_elements):
            if i > 0:
                target_field += "."
            target_field += key_path_element

        response_item: Optional[dict] = self._get_item_by_primary_key(
            key_name=key_name, key_value=key_value, fields_to_get=[target_field]
        ).item
        for i, key_path_element in enumerate(target_path_elements):
            if i + 1 > num_keys_to_navigation_into:
                break

            response_item = response_item.get(key_path_element, None)
            if response_item is None:
                # If the response_item is None, it means that one key has not been found,
                # so we need to break the loop in order to try to call get on a None object,
                # and then we will return the response_item, so we will return None.
                break
        return response_item

    def _get_value_in_path_target(self, key_name: str, key_value: str, target_path_elements: Dict[str, type]) -> Optional[any]:
        return self._get_data_in_path_target(
            key_name=key_name, key_value=key_value,
            target_path_elements=target_path_elements,
            num_keys_to_navigation_into=len(target_path_elements)
        )

    def _get_item_in_path_target(self, key_name: str, key_value: str, target_path_elements: Dict[str, type]) -> Optional[dict]:
        return self._get_data_in_path_target(
            key_name=key_name, key_value=key_value,
            target_path_elements=target_path_elements,
            num_keys_to_navigation_into=len(target_path_elements) - 1
        )

    @staticmethod
    def _add_to_filter_expression(expression, condition):
        if expression is None:
            return condition
        return expression & condition

    @staticmethod
    def _fields_to_get_to_expressions(fields_to_get: List[str]) -> dict:
        output_kwargs = {}
        expression_attribute_names = {}
        projection_expression = ""

        # In DynamoDB, when trying to get some fields, certain type of values (like values with - in them,
        # or big numbers like an UUID), can not work and provoke in error while executing the query. In
        # order to fix that, we need to pass the variable name in the ExpressionAttributeNames instead of
        # putting it directly in the ProjectionExpression. Yet, when using a map path like myMap.data.item1
        # we must declare each path element (myMap, data, item1) as separate ExpressionAttributeNames,
        # otherwise it will not work. So, we define an 'id' for each path of each attribute name
        # (f"#f{i_field}_{i_path}") we add our path as the dict value, and then we build the condition
        # expression to use the ids of our attributes names.
        for i_field, field in enumerate(fields_to_get):
            field: str  # PyCharm did not recognized the field variable as a string ;)
            path_elements = field.split(".")

            for i_path, path in enumerate(path_elements):
                current_field_path_expression_name = f"#f{i_field}_{i_path}"
                expression_attribute_names[current_field_path_expression_name] = path
                projection_expression += current_field_path_expression_name
                if i_path + 1 < len(path_elements):
                    projection_expression += "."

            if i_field + 1 < len(fields_to_get):
                projection_expression += ", "

        if len(expression_attribute_names) > 0:
            output_kwargs["ExpressionAttributeNames"] = expression_attribute_names
        if projection_expression.replace(" ", "") != "":
            output_kwargs["ProjectionExpression"] = projection_expression

        return output_kwargs





from inoft_vocal_engine import SpeechsList, Speech

MSGS_START_FACT = SpeechsList(id="startFactPhrase", speechs=[
    Speech().add_text("Did you know").set_prob(1),
    Speech().add_text("Have you heard").set_prob(1),
    Speech().add_text("Did you suspected").set_prob(0.5)
])

MSGS_END_FACT = SpeechsList(id="endFactPhrase", speechs=[
    Speech().add_text("Don't you found it wild ?").set_prob(0.5),
    Speech().add_text("Don't you found it crazy ?").set_prob(1),
    Speech().add_text("Crazy, right ?").set_prob(1),
    Speech().add_text("Its not crazy to you ? It is for me !").set_prob(1),
])

MSGS_NEW_FACT = SpeechsList(id="newFactCallToAction", speechs=[
    Speech().add_text("Ready for a new fact ?").set_prob(1),
    Speech().add_text("Do you want another fact ?").set_prob(1),
])

MSGS_FACTS = SpeechsList(id="mainFacts", speechs=[
    Speech().add_text("that the Inoft Vocal Framework was the first package from Robinson Labourdette ? He did not even knew how to publish it !").set_prob(1),
    Speech().add_text("that for simpler use of the framework and the processing of the requests, a new object type, the SafeDict, has been created ?").set_prob(1),
    Speech().add_text("that i'm starting to be out of inspiration ?").set_prob(0.1)
])



import os
from inoft_vocal_framework.plugins.plugin_base import PluginCodeGenerationBase


class Core(PluginCodeGenerationBase):
    from inoft_vocal_framework.botpress_integration import GeneratorCore

    def execute(self, generator_core: GeneratorCore):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        generator_core.templates.say_action_template = os.path.join(current_dir, "say_action.tem")

        bucket_name = "test-vocal-export-inoft"
        bucket_region = "eu-west-3"

        for say_action in generator_core.say_actions:
            say_action.code = f"https://{bucket_name}.s3.{bucket_region}.amazonaws.com/{say_action.message_id_to_say}/compressed_final_render.mp3"



from typing import Any, Tuple, Optional, List
from pydantic import validate_arguments
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.web_interface.auth.models import AccountData


class AccountResources:
    @validate_arguments
    def __init__(self, engine_resources: Any):
        self.engine_resources: EngineResources = engine_resources
        # self.account_username = account_username
        # self._account_id = None
        self.is_logged_in = False
        self.account_id = None
        self.account_username = None

    """
    @property
    def account_id(self) -> str:
        if self._account_id is None:
            account_data, account_been_found = self.engine_resources.accounts_data_dynamodb_client.get_account_by_username(
               username=self.account_username, fields_to_get=["accountId"],
            )
            if account_been_found is True:
                account_data = AccountData(**account_data)
                self._account_id = account_data.accountId
        return self._account_id
    """

    def _is_valid_auth_token_and_remove_old_tokens(self, account_id: str, auth_token: str) -> bool:
        account_data = self.engine_resources.accounts_data_dynamodb_client.remove_old_auth_tokens(
            account_id=account_id, fields_to_get=["accountId", "authTokens", "username"]
        )
        for valid_auth_token_item in account_data.authTokens:
            if auth_token == valid_auth_token_item.token:
                self.is_logged_in = True
                self.account_id = account_data.accountId
                self.account_username = account_data.username
                return True
        return False

    def login(self, username_or_email: str, encrypted_password: str) -> Tuple[bool, bool, Optional[str]]:
        if self.is_logged_in is False:
            account_has_been_logged_in = False
            new_auth_token_value = None
            fields_to_get = ["accountId", "username", "encryptedPassword"]

            account_data = self.engine_resources.accounts_data_dynamodb_client.get_account_by_email(
                email=username_or_email, fields_to_get=fields_to_get
            )
            if account_data is None:
                account_data = self.engine_resources.accounts_data_dynamodb_client.get_account_by_username(
                    username=username_or_email, fields_to_get=fields_to_get
                )

            if account_data is not None:
                print(account_data)
                account_data = AccountData(**account_data)

                if account_data.encryptedPassword is not None:
                    print(f"encrypted_password from the user : {encrypted_password}\n"
                          f"encrypted_password of the account : {account_data.encryptedPassword}")

                    if encrypted_password.lower() == account_data.encryptedPassword.lower():
                        # We lower the cases of both the passwords, to avoid issues of un-important conversion of the data
                        print("Password matched")
                        self.engine_resources.accounts_data_dynamodb_client.remove_old_auth_tokens(
                            account_id=account_data.accountId, fields_to_get=[]
                        )
                        new_auth_token_value = self.engine_resources.accounts_data_dynamodb_client.add_auth_token(account_id=account_data.accountId)
                        account_has_been_logged_in = True

                        self.is_logged_in = True
                        self.account_id = account_data.accountId
                        self.account_username = account_data.username
                    else:
                        print("Password did not match")
                else:
                    print(f"WARNING ! No password field has been found for user {username_or_email} with user_data : {account_data}")

                return True, account_has_been_logged_in, new_auth_token_value
            else:
                return False, account_has_been_logged_in, new_auth_token_value

    def login_with_cookies(self, cookies: Any) -> bool:
        from werkzeug.datastructures import ImmutableMultiDict
        if isinstance(cookies, (ImmutableMultiDict, dict)):
            account_id = cookies.get("lastLoggedWithAccountId")
            auth_token = cookies.get("authToken")

            if account_id is not None and isinstance(account_id, str) and auth_token is not None and isinstance(auth_token, str):
                return self._is_valid_auth_token_and_remove_old_tokens(account_id=account_id, auth_token=auth_token)
            else:
                return False
        else:
            raise Exception(f"Cookies must be of instance {ImmutableMultiDict} or {dict}")

    def check_for_permissions(self, project_owner_account_username: str, permissions_expressions: List[str]) -> Tuple[Optional[str], bool]:
        if self.is_logged_in is True and project_owner_account_username == self.account_username:
            print(f"Account being checked for permissions on a project that he owns. "
                  f"Permissions check not executed, and permissions approved."
                  f"\n  --project_owner_account_username:{project_owner_account_username}"
                  f"\n  --permissions_expressions(permissions_to_check):{permissions_expressions}")
            return self.account_id, True
        else:
            response = self.engine_resources.accounts_data_dynamodb_client.check_permissions_on_account_project(
                project_owner_account_username=project_owner_account_username,
                requester_account_id=self.account_id,
                permissions_to_check=permissions_expressions,
                fields_to_get=["accountId"]
            )
            if len(response.items) > 0:
                project_owner_account_data = AccountData(**response.items[0])
                print(f"Permissions approved for requester account for a project of another account."
                      f"\n  --project_owner_account_username:{project_owner_account_username}"
                      f"\n  --account_id(requester_account):{self.account_id}"
                      f"\n  --permissions_expressions(permissions_to_check):{permissions_expressions}"
                      f"\n  --project_owner_account_data(validated response data):{project_owner_account_data}")
                return project_owner_account_data.accountId, True
            else:
                print(f"Permissions not approved for requester account for a project of another account."
                      f"\n  --project_owner_account_username:{project_owner_account_username}"
                      f"\n  --account_id(requester_account):{self.account_id}"
                      f"\n  --permissions_expressions(permissions_to_check):{permissions_expressions}")
                return None, False



from flask import Blueprint, render_template


permissions_blueprint = Blueprint("permissions", __name__, template_folder="templates")

@permissions_blueprint.route("/permissions/missing")
def permissions_missing():
    return render_template("permissions/missing.html")



from enum import Enum

from pydantic import BaseModel
from typing import List, Optional, Any, Dict

from pydantic.dataclasses import dataclass


class AlexaModelConfiguration(BaseModel):
    class FallbackIntentSensitivityModel(BaseModel):
        class LevelOptions(str, Enum):
            HIGH = "HIGH"
            MEDIUM = "MEDIUM"
            LOW = "LOW"
        level: LevelOptions
    fallbackIntentSensitivity: Optional[FallbackIntentSensitivityModel] = None

class AlexaModelSchema(BaseModel):
    class InteractionModelModel(BaseModel):
        class LanguageModelModel(BaseModel):
            invocationName: str
            modelConfiguration: Optional[AlexaModelConfiguration] = None
        languageModel: LanguageModelModel

        class IntentModel(BaseModel):
            class SlotModel(BaseModel):
                name: str
                type: str
                samples: List[str]
            name: str
            samples: List[str]
            slots: List[SlotModel]
        intents: List[IntentModel]

    interactionModel: InteractionModelModel


class InoftVocalEngineModelSchema(BaseModel):
    invocationName: str
    LanguageModel: Optional[str] = None

    class IntentModel(BaseModel):
        class SlotModel(BaseModel):
            name: str
            type: str
            samples: Optional[List[str]] = list()
        intentId: str
        name: str
        samples: Optional[List[str]] = list()
        slots: Optional[List[SlotModel]] = list()
    intents: Optional[Dict[str, IntentModel]] = dict()

    class AlexaSpecific(BaseModel):
        modelConfiguration: Optional[AlexaModelConfiguration] = None
    alexaSpecific: Optional[AlexaSpecific] = None

    class PublishingModel(BaseModel):
        class AlexaModel(BaseModel):
            class LocaleModel(BaseModel):
                skillName: str
            locales: Dict[str, LocaleModel]
        alexa: AlexaModel
    publishing: PublishingModel

class InoftVocalEngineModelSchemaDatabaseComplete(InoftVocalEngineModelSchema):
    accountProjectId: str = None


def inoft_vocal_engine_model_schema_to_alexa_model_schema(inoft_vocal_engine_model_instance: InoftVocalEngineModelSchema) -> AlexaModelSchema:
    intents: List[AlexaModelSchema.InteractionModelModel.IntentModel] = list()
    for src_intent in inoft_vocal_engine_model_instance.intents.values():
        intents.append(AlexaModelSchema.InteractionModelModel.IntentModel(
            name=src_intent.name,
            samples=src_intent.samples,
            slots=src_intent.slots
        ))

    output_alexa_model_schema = AlexaModelSchema(
        interactionModel=AlexaModelSchema.InteractionModelModel(
            languageModel=AlexaModelSchema.InteractionModelModel.LanguageModelModel(
                invocationName=inoft_vocal_engine_model_instance.invocationName,
                modelConfiguration=inoft_vocal_engine_model_instance.alexaSpecific.modelConfiguration
            ), intents=intents
        )
    )
    return output_alexa_model_schema

# todo: save the alexa model schema in json format and store it in S3 ? I have no fucking clue, i'm too tired, i just want to sleep --'


if __name__ == "__main__":
    m = InoftVocalEngineModelSchema(invocationName="test", LanguageModel="test", intents={},
                                    alexaSpecific={"modelConfiguration": {
                                        "fallbackIntentSensitivity": {
                                          "level": "LOW"
                                        }
                                    }})
    print(m)
    print(inoft_vocal_engine_model_schema_to_alexa_model_schema(m))

"""
Alexa :
    {
  "interactionModel": {
    "languageModel": {
      "invocationName": "my space facts",
      "modelConfiguration": {
        "fallbackIntentSensitivity": {
          "level": "LOW"
        }
      },
      "intents": [
        {
          "name": "GetTravelTime",
          "slots": [
            {
              "name": "DepartingPlanet",
              "type": "Planet",
              "samples": [
                "I'm starting from {DepartingPlanet} ",
                "{DepartingPlanet} ",
                "I'm going from {DepartingPlanet} to {ArrivingPlanet} "
              ]
            },
            {
              "name": "ArrivingPlanet",
              "type": "Planet",
              "samples": [
                "I'm going to {ArrivingPlanet} ",
                "{ArrivingPlanet} "
              ]
            }
          ],
          "samples": [
            "calculate travel time",
            "how long does it take to travel from {DepartingPlanet} to {ArrivingPlanet} "
          ]
        }
      ],
      "types": [
        {
          "name": "Planet",
          "values": [
            {
              "name": {
                "value": "Mercury"
              }
            },
            {
              "name": {
                "value": "Venus"
              }
            },
            {
              "name": {
                "value": "Earth"
              }
            },
            {
              "name": {
                "value": "Mars"
              }
            },
            {
              "name": {
                "value": "Jupiter"
              }
            },
            {
              "name": {
                "value": "Saturn"
              }
            },
            {
              "name": {
                "value": "Uranus"
              }
            },
            {
              "name": {
                "value": "Neptune"
              }
            },
            {
              "name": {
                "value": "Pluto"
              }
            }
          ]
        }
      ]
    },
    "dialog": {
      "intents": [
        {
          "name": "GetTravelTime",
          "confirmationRequired": false,
          "prompts": {},
          "slots": [
            {
              "name": "DepartingPlanet",
              "type": "Planet",
              "confirmationRequired": false,
              "elicitationRequired": true,
              "prompts": {
                "elicitation": "Elicit.Intent-GetTravelTime.IntentSlot-DepartingPlanet"
              },
              "validations": [
                {
                  "type": "isNotInSet",
                  "prompt": "Slot.Validation.596358663326.282490667310.1526107495625",
                  "values": [
                    "the sun",
                    "sun",
                    "our sun"
                  ]
                },
                {
                  "type": "hasEntityResolutionMatch",
                  "prompt": "Slot.Validation.596358663326.282490667310.1366622834897"
                }
              ]
            },
            {
              "name": "ArrivingPlanet",
              "type": "Planet",
              "confirmationRequired": false,
              "elicitationRequired": true,
              "prompts": {
                "elicitation": "Elicit.Intent-GetTravelTime.IntentSlot-ArrivingPlanet"
              }
            }
          ]
        }
      ],
      "delegationStrategy": "ALWAYS"
    },
    "prompts": [
      {
        "id": "Elicit.Intent-GetTravelTime.IntentSlot-DepartingPlanet",
        "variations": [
          {
            "type": "PlainText",
            "value": "Which planet do you want to start from?"
          }
        ]
      },
      {
        "id": "Elicit.Intent-GetTravelTime.IntentSlot-ArrivingPlanet",
        "variations": [
          {
            "type": "PlainText",
            "value": "Which planet do you want to travel to?"
          }
        ]
      },
      {
        "id": "Slot.Validation.596358663326.282490667310.1526107495625",
        "variations": [
          {
            "type": "PlainText",
            "value": "I can't answer this question about the sun, only planets. Please tell me a planet."
          },
          {
            "type": "PlainText",
            "value": "While the sun is the center of our solar system, it is not a planet. Please tell me a planet."
          }
        ]
      },
      {
        "id": "Slot.Validation.596358663326.282490667310.1366622834897",
        "variations": [
          {
            "type": "PlainText",
            "value": "{DepartingPlanet} is not a planet. Please tell me one of the nine planets in our solar system. "
          },
          {
            "type": "PlainText",
            "value": "I don't recognize {DepartingPlanet} as a planet in our solar system. Please tell me a planet."
          }
        ]
      }
    ]
  }
}
"""


import os
from typing import List

from inoft_vocal_framework.plugins.plugin_base import PluginBase, PluginCodeGenerationBase
from inoft_vocal_framework.skill_settings.skill_settings import Settings


def plugins_load(settings: Settings) -> List[PluginCodeGenerationBase]:
    plugins_dir_path = os.path.dirname(os.path.abspath(__file__))
    cores_instances_of_code_generation_plugins: List[PluginCodeGenerationBase] = list()
    # We separate the different type of plugins in different lists, because different plugins will
    # not be executed at the same time, or need to same arguments in their execute function.

    # The activated plugins variable is a list containing with the dir name of all the activated plugins
    for plugin_dir_name in settings.plugins.activated_plugings:
        current_plugin_dirpath = os.path.join(plugins_dir_path, plugin_dir_name)
        if not os.path.isdir(current_plugin_dirpath):
            raise Exception(f"The plugin named {plugin_dir_name} has not been found at {current_plugin_dirpath}. Did you type the name right ?")
        else:
            current_plugin_core_filepath = os.path.join(current_plugin_dirpath, "core.py")
            if not os.path.isfile(current_plugin_core_filepath):
                raise Exception(f"The plugin named {plugin_dir_name} has been found at {current_plugin_dirpath}.\n"
                                f"But no file named core.py has been found at the root of this folder.\n"
                                f"Maybe try to re-download the plugin, or it is not compatible with your version of the Inoft Vocal Engine.")
            else:
                import importlib
                import importlib.util
                spec = importlib.util.spec_from_file_location("plugin_core", current_plugin_core_filepath)
                current_plugin_core_module = importlib.util.module_from_spec(spec=spec)
                spec.loader.exec_module(current_plugin_core_module)

                vars_current_plugin_core_module = vars(current_plugin_core_module)
                if "Core" not in vars_current_plugin_core_module.keys():
                    raise Exception(f"The class named Core is missing from the plugin core file at {current_plugin_core_filepath}")
                else:
                    current_plugin_core_class_type = vars_current_plugin_core_module["Core"]

                    from inspect import getmro
                    current_plugin_core_class_parents = getmro(current_plugin_core_class_type)
                    if PluginBase not in current_plugin_core_class_parents:
                        raise Exception(f"The Core class of plugin {current_plugin_dirpath} found at {current_plugin_core_filepath} "
                                        f"was not is not using one of the valid type of plugins parent classes. "
                                        f"It had the following classes : {current_plugin_core_class_parents}")
                    else:
                        if PluginCodeGenerationBase in current_plugin_core_class_parents:
                            cores_instances_of_code_generation_plugins.append(current_plugin_core_class_type())
                        else:
                            raise Exception(f"No valid type of Plugin parent class had been found for the Core class found at {current_plugin_core_filepath}")

    print(cores_instances_of_code_generation_plugins)





from setuptools import setup, find_packages

setup(
    name="inoftvocal",
    version="0.90.5.3",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["PyYAML", "pydantic", "boto3", "click", "inflect", "discord.py"],
    entry_points={
        "console_scripts": [
            "inoft = inoft_vocal_framework.cli.cli_index:cli",
            "inoftvocal = inoft_vocal_framework.cli.cli_index:cli",
        ],
    },
    url="https://github.com/Robinson04/inoft_vocal_framework",
    license="MIT",
    author="Inoft",
    author_email="robinson@inoft.com",
    description="Create advanced cross-platform skills for Alexa, Google Assistant and Samsung Bixby",
)




import time
from asyncio import Future
from concurrent.futures.thread import ThreadPoolExecutor
import boto3
from typing import List

from inoft_vocal_engine.cloud_providers.aws.policies import ATTACH_POLICY, ASSUME_POLICY


class AwsCoreClients:
    CLIENT_BOTO_SESSION = 10000
    CLIENT_S3 = 10001
    CLIENT_DYNAMODB = 10002
    CLIENT_LAMBDA = 10003
    CLIENT_API_GATEWAY = 10004
    CLIENT_API_GATEWAY_V2 = 10005
    CLIENT_EVENTS = 10006
    CLIENT_IAM = 10007
    RESOURCE_IAM = 10008
    CLIENT_STS = 10009

    def __init__(self, clients_to_load: List[int]):
        self._boto_session, self._s3_client, self._lambda_client, self._dynamodb_client, \
        self._api_gateway_client, self._api_gateway_v2_client, self._events_client, \
        self._iam_client, self._iam_resource, self._sts_client = None, None, None, None, None, None, None, None, None, None

        self.attach_policy = ATTACH_POLICY
        self.assume_policy = ASSUME_POLICY

        self.pool = ThreadPoolExecutor(len(clients_to_load))
        # We cannot use asyncio for the initialization of the client, because currently the http
        # requests made by boto3 are blocking http request made with the default urllib library.
        # See (https://www.mathewmarcus.com/blog/asynchronous-aws-api-requests-with-asyncio.html)

        futures: List[Future] = list()
        if self.CLIENT_BOTO_SESSION in clients_to_load:
            futures.append(self.pool.submit(self._create_boto_session))
        if self.CLIENT_S3 in clients_to_load:
            futures.append(self.pool.submit(self._create_s3_client))
        if self.CLIENT_DYNAMODB in clients_to_load:
            futures.append(self.pool.submit(self._create_dynamodb_client))
        if self.CLIENT_LAMBDA in clients_to_load:
            futures.append(self.pool.submit(self._create_lambda_client))
        if self.CLIENT_API_GATEWAY in clients_to_load:
            futures.append(self.pool.submit(self._create_api_gateway_client))
        if self.CLIENT_API_GATEWAY_V2 in clients_to_load:
            futures.append(self.pool.submit(self._create_api_gateway_v2_client))
        if self.CLIENT_EVENTS in clients_to_load:
            futures.append(self.pool.submit(self._create_events_client))
        if self.CLIENT_IAM in clients_to_load:
            futures.append(self.pool.submit(self._create_iam_client))
        if self.RESOURCE_IAM in clients_to_load:
            futures.append(self.pool.submit(self._create_iam_resource))
        if self.CLIENT_STS in clients_to_load:
            futures.append(self.pool.submit(self._create_sts_client))

        for future in futures:
            while not future.done():
                time.sleep(0.01)

    def _raise_client_not_loaded(self, client_variable_name: str):
        raise Exception(f"{client_variable_name} was not in the clients_to_load list.")

    @property
    def boto_session(self) -> boto3.Session:
        if self._boto_session is None:
            self._raise_client_not_loaded("CLIENT_BOTO_SESSION")
        return self._boto_session
    
    def _create_boto_session(self):
        self._boto_session = boto3.Session()
        
    @property
    def s3_client(self) -> boto3.client:
        if self._s3_client is None:
            self._raise_client_not_loaded("CLIENT_S3")
        return self._s3_client

    def _create_s3_client(self):
        self._s3_client = boto3.client("s3")

    @property
    def dynamodb_client(self) -> boto3.client:
        if self._dynamodb_client is None:
            self._raise_client_not_loaded("CLIENT_DYNAMODB")
        return self._dynamodb_client

    def _create_dynamodb_client(self):
        self._dynamodb_client = boto3.client("dynamodb")
        
    @property
    def lambda_client(self) -> boto3.client:
        if self._lambda_client is None:
            self._raise_client_not_loaded("CLIENT_LAMBDA")
        return self._lambda_client

    def _create_lambda_client(self):
        self._lambda_client = boto3.client("lambda")

    @property
    def api_gateway_client(self) -> boto3.client:
        if self._api_gateway_client is None:
            self._raise_client_not_loaded("CLIENT_API_GATEWAY")
        return self._api_gateway_client

    def _create_api_gateway_client(self):
        self._api_gateway_client = boto3.client("apigateway")

    @property
    def api_gateway_v2_client(self) -> boto3.client:
        if self._api_gateway_v2_client is None:
            self._raise_client_not_loaded("CLIENT_API_GATEWAY_V2")
        return self._api_gateway_v2_client

    def _create_api_gateway_v2_client(self):
        self._api_gateway_v2_client = boto3.client("apigatewayv2")
        
    @property
    def events_client(self) -> boto3.client:
        if self._events_client is None:
            self._raise_client_not_loaded("CLIENT_EVENTS")
        return self._events_client

    def _create_events_client(self):
        self._events_client = boto3.client("events")
        
    @property
    def iam_client(self) -> boto3.client:
        if self._iam_client is None:
            self._raise_client_not_loaded("CLIENT_IAM")
        return self._iam_client

    def _create_iam_client(self):
        self._iam_client = boto3.client("iam")
        # todo: check if the iam from the boto_session is equivalent to the classical iam_client
        
    @property
    def iam_resource(self) -> boto3.client:
        if self._iam_resource is None:
            self._raise_client_not_loaded("RESOURCE_IAM")
        return self._iam_resource

    def _create_iam_resource(self):
        self._iam_resource = boto3.resource("iam")

    @property
    def sts_client(self) -> boto3.client:
        if self._sts_client is None:
            self._raise_client_not_loaded("STS_CLIENT")
        return self._sts_client

    def _create_sts_client(self):
        self._sts_client = boto3.client("sts")



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)
Copyright (c) 2015-2019 Rapptz
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import time
import wave
import select
import socket
import audioop
import logging
import threading
import traceback

from inoft_vocal_framework.platforms_handlers.discord.lib_customs import rtp
from inoft_vocal_framework.platforms_handlers.discord.lib_customs.opus import Decoder, BufferedDecoder
from inoft_vocal_framework.platforms_handlers.discord import VoiceClient
from discord.errors import DiscordException

try:
    import nacl.secret
    from nacl.exceptions import CryptoError
except ImportError:
    nacl = None
    # Just to avoid type checking errors

log = logging.getLogger(__name__)

__all__ = [
    DiscordException,
    'AudioSink',
    'WaveSink',
    'PCMVolumeTransformerFilter',
    'ConditionalFilter',
    'TimedFilter',
    'UserFilter',
    'SinkExit'
]

class SinkExit(DiscordException):
    """A signal type exception (like ``GeneratorExit``) to raise in a Sink's write() method to stop it.
    TODO: make better words
    Parameters
    -----------
    drain: :class:`bool`
        ...
    flush: :class:`bool`
        ...
    """

    def __init__(self, *, drain=True, flush=False):
        self.kwargs = kwargs

class AudioSink:
    def __del__(self):
        self.cleanup()

    def write(self, data):
        raise NotImplementedError

    def wants_opus(self):
        return False

    def cleanup(self):
        pass

    def pack_data(self, data, user=None, packet=None):
        return VoiceData(data, user, packet)  # is this even necessary?

class WaveSink(AudioSink):
    def __init__(self, destination):
        self._file = wave.open(destination, 'wb')
        self._file.setnchannels(Decoder.CHANNELS)
        self._file.setsampwidth(Decoder.SAMPLE_SIZE//Decoder.CHANNELS)
        self._file.setframerate(Decoder.SAMPLING_RATE)

    def write(self, data):
        self._file.writeframes(data.data)

    def cleanup(self):
        try:
            self._file.close()
        except:
            pass

class PCMVolumeTransformerFilter(AudioSink):
    def __init__(self, destination, volume=1.0):
        if not isinstance(destination, AudioSink):
            raise TypeError('expected AudioSink not {0.__class__.__name__}.'.format(destination))

        if destination.wants_opus():
            raise ClientException('AudioSink must not request Opus encoding.')

        self.destination = destination
        self.volume = volume

    @property
    def volume(self):
        """Retrieves or sets the volume as a floating point percentage (e.g. 1.0 for 100%)."""
        return self._volume

    @volume.setter
    def volume(self, value):
        self._volume = max(value, 0.0)

    def write(self, data):
        data = audioop.mul(data.data, 2, min(self._volume, 2.0))
        self.destination.write(data)

# I need some sort of filter sink with a predicate or something
# Which means I need to sort out the write() signature issue
# Also need something to indicate a sink is "done", probably
# something like raising an exception and handling that in the write loop
# Maybe should rename some of these to Filter instead of Sink

class ConditionalFilter(AudioSink):
    def __init__(self, destination, predicate):
        self.destination = destination
        self.predicate = predicate

    def write(self, data):
        if self.predicate(data):
            self.destination.write(data)

class TimedFilter(ConditionalFilter):
    def __init__(self, destination, duration, *, start_on_init=False):
        super().__init__(destination, self._predicate)
        self.duration = duration
        if start_on_init:
            self.start_time = self.get_time()
        else:
            self.start_time = None
            self.write = self._write_once

    def _write_once(self, data):
        self.start_time = self.get_time()
        super().write(data)
        self.write = super().write

    def _predicate(self, data):
        return self.start_time and self.get_time() - self.start_time < self.duration

    def get_time(self):
        return time.time()

class UserFilter(ConditionalFilter):
    def __init__(self, destination, user):
        super().__init__(destination, self._predicate)
        self.user = user

    def _predicate(self, data):
        return data.user == self.user

# rename 'data' to 'payload'? or 'opus'? something else?
class VoiceData:
    __slots__ = ('data', 'user', 'packet')

    def __init__(self, data, user, packet):
        self.data = data
        self.user = user
        self.packet = packet

class AudioReader(threading.Thread):
    def __init__(self, sink: AudioSink, client: VoiceClient, *, after=None):
        super().__init__(daemon=True)
        self.sink = sink
        self.client = client
        self.after = after

        if after is not None and not callable(after):
            raise TypeError('Expected a callable for the "after" parameter.')

        self.after = after

        self.box = nacl.secret.SecretBox(bytes(client.secret_key))
        self.decrypt_rtp = getattr(self, '_decrypt_rtp_' + client._mode)
        self.decrypt_rtcp = getattr(self, '_decrypt_rtcp_' + client._mode)

        self._current_error = None
        self._end = threading.Event()
        self._decoder_lock = threading.Lock()

        self.decoder = BufferedDecoder(self)
        self.decoder.start()

        # TODO: inject sink functions

    @property
    def connected(self):
        return self.client._connected

    def _decrypt_rtp_xsalsa20_poly1305(self, packet):
        nonce = bytearray(24)
        nonce[:12] = packet.header
        result = self.box.decrypt(bytes(packet.data), bytes(nonce))

        if packet.extended:
            offset = packet.update_ext_headers(result)
            result = result[offset:]

        return result

    def _decrypt_rtcp_xsalsa20_poly1305(self, data):
        nonce = bytearray(24)
        nonce[:8] = data[:8]
        result = self.box.decrypt(data[8:], bytes(nonce))

        return data[:8] + result

    def _decrypt_rtp_xsalsa20_poly1305_suffix(self, packet):
        nonce = packet.data[-24:]
        voice_data = packet.data[:-24]
        result = self.box.decrypt(bytes(voice_data), bytes(nonce))

        if packet.extended:
            offset = packet.update_ext_headers(result)
            result = result[offset:]

        return result

    def _decrypt_rtcp_xsalsa20_poly1305_suffix(self, data):
        nonce = data[-24:]
        header = data[:8]
        result = self.box.decrypt(data[8:-24], nonce)

        return header + result

    def _decrypt_rtp_xsalsa20_poly1305_lite(self, packet):
        nonce = bytearray(24)
        nonce[:4] = packet.data[-4:]
        voice_data = packet.data[:-4]
        result = self.box.decrypt(bytes(voice_data), bytes(nonce))

        if packet.extended:
            offset = packet.update_ext_headers(result)
            result = result[offset:]

        return result

    def _decrypt_rtcp_xsalsa20_poly1305_lite(self, data):
        nonce = bytearray(24)
        nonce[:4] = data[-4:]
        header = data[:8]
        result = self.box.decrypt(data[8:-4], bytes(nonce))

        return header + result

    def _reset_decoders(self, *ssrcs):
        self.decoder.reset(*ssrcs)

    def _stop_decoders(self, **kwargs):
        self.decoder.stop(**kwargs)

    def _ssrc_removed(self, ssrc):
        # An user has disconnected but there still may be
        # packets from them left in the buffer to read
        # For now we're just going to kill the decoder and see how that works out
        # I *think* this is the correct way to do this
        # Depending on how many leftovers I end up with I may reconsider

        self.decoder.drop_ssrc(ssrc) # flush=True?

    def _get_user(self, packet):
        _, user_id = self.client._get_ssrc_mapping(ssrc=packet.ssrc)
        # may need to change this for calls or something
        return self.client.guild.get_member(user_id)

    def _write_to_sink(self, pcm, opus, packet):
        try:
            data = opus if self.sink.wants_opus() else pcm
            user = self._get_user(packet)
            self.sink.write(VoiceData(data, user, packet))
            # TODO: remove weird error handling in favor of injected functions
        except SinkExit as e:
            log.info("Shutting down reader thread %s", self)
            self.stop()
            self._stop_decoders(**e.kwargs)
        except:
            traceback.print_exc()
            # insert optional error handling here

    def _set_sink(self, sink):
        with self._decoder_lock:
            self.sink = sink
        # if i were to fire a sink change mini-event it would be here

    def _do_run(self):
        while not self._end.is_set():
            if not self.connected.is_set():
                self.connected.wait()

            is_socket_ready_for_reading, _, has_socket_experienced_exception = select.select(
                rlist=[self.client.socket], wlist=[], xlist=[self.client.socket], timeout=0.01)
            # todo: test if this works on windows
            # _ is receiving is_socket_ready_for_writing, which we do not give a damn about,
            # and anyway, we did not even send the socket to the writing list (wlist)

            if not is_socket_ready_for_reading:
                if has_socket_experienced_exception:
                    print("Socket error")
                continue

            try:
                audio_raw_data = self.client.socket.recv(4096)

            except socket.error as e:
                t0 = time.time()

                if e.errno == 10038:  # ENOTSOCK
                    continue

                log.exception("Socket error in reader thread ")
                print(f"Socket error in reader thread: {e} {t0}")

                with self.client._connecting:
                    timed_out = self.client._connecting.wait(20)

                if not timed_out:
                    raise
                elif self.client.is_connected():
                    print(f"Reconnected in {time.time()-t0:.4f}s")
                    continue
                else:
                    raise

            try:
                packet = None
                if not rtp.is_rtcp(audio_raw_data):
                    packet = rtp.decode(audio_raw_data)
                    packet.decrypted_data = self.decrypt_rtp(packet)
                else:
                    packet = rtp.decode(self.decrypt_rtcp(audio_raw_data))
                    if not isinstance(packet, rtp.ReceiverReportPacket):
                        print(packet)

                        # TODO: Fabricate and send SenderReports and see what happens

                    self.decoder.feed_rtcp(packet)
                    continue

            except CryptoError:
                log.exception("CryptoError decoding packet %s", packet)
                continue

            except:
                log.exception("Error unpacking packet")
                traceback.print_exc()

            else:
                if packet.ssrc not in self.client._ssrcs:
                    log.debug("Received packet for unknown ssrc %s", packet.ssrc)

                self.decoder.feed_rtp(packet)

    def stop(self):
        self._end.set()

    def run(self):
        try:
            self._do_run()
        except socket.error as exc:
            self._current_error = exc
            self.stop()
        except Exception as exc:
            traceback.print_exc()
            self._current_error = exc
            self.stop()
        finally:
            self._stop_decoders()
            try:
                self.sink.cleanup()
            except:
                log.exception("Error during sink cleanup")
                # Testing only
                traceback.print_exc()

            self._call_after()

    def _call_after(self):
         if self.after is not None:
            try:
                self.after(self._current_error)
            except Exception:
                log.exception('Calling the after function failed.')

    def is_listening(self):
        return not self._end.is_set()


if __name__ == "__main__":
    AudioReader()


from dataclasses import field
from typing import Optional, List, Dict, Any
from pydantic import BaseModel, Field, validator
from pydantic.dataclasses import dataclass
from inoft_vocal_engine.inoft_vocal_markup.deserializer import DialogueLine
from inoft_vocal_engine.models.actions import BaseAction
from inoft_vocal_engine.models.conditions import BaseCondition


class ContentElement(BaseModel):
    id: str = None
    created_by: str = None
    created_on: str = None
    modified_on: str = None
    crude_text: str = None
    character_names: list = None
    dialogues_lines: Optional[List[DialogueLine]] = None

class DatabaseTransitionConditionItem(BaseModel):
    transitionId: str = None
    condition: Optional[dict] = None
    sourceNodeId: Optional[str] = None
    targetNodeId: Optional[str] = None

    """@validator("condition")
    def get_appropriate_condition(cls, value):
        from inoft_vocal_engine.models.conditions import get_appropriate_condition
        return get_appropriate_condition(condition_data=value)"""

    def to_database(self) -> dict:
        return self.dict(exclude={"transitionId", "sourceNodeId"})

class ProcessedTransitionCondition(BaseModel):
    transitionId: str
    condition: Optional[BaseCondition] = None
    sourceNodeId: Optional[str] = None
    targetNodeId: Optional[str] = None

    @classmethod
    def from_database_item(cls, database_transition_condition_item: DatabaseTransitionConditionItem):
        from inoft_vocal_engine.models.conditions import get_appropriate_condition
        return cls(
            transitionId=database_transition_condition_item.transitionId,
            condition=get_appropriate_condition(database_transition_condition_item.condition),
            sourceNodeId=database_transition_condition_item.sourceNodeId,
            targetNodeId=database_transition_condition_item.targetNodeId
        )

class ProcessedNodeElement(BaseModel):
    nodeId: str = None
    nodeName: str = "Unnamed"
    actions: Optional[List[BaseAction]] = list()
    transitions: Optional[List[ProcessedTransitionCondition]] = list()
    temp_on_receive: Optional[list] = list()
    nodeType: str = None
    positionX: int = 0
    positionY: int = 0
    # todo: make stricter variables validation

class NodeElementDataReceivedFromUser(BaseModel):
    nodeId: str
    nodeName: Optional[str] = None
    actions: Optional[List[dict]]
    transitions: Optional[Dict[str, DatabaseTransitionConditionItem]]
    temp_on_receive: Optional[list] = None
    nodeType: Optional[str] = None
    positionX: Optional[int] = None
    positionY: Optional[int] = None

class NodeDatabaseItem(BaseModel):
    accountId: str
    accountProjectId: str
    accountProjectInstanceId: str
    nodeId: str
    nodeName: str
    positionX: int = 0
    positionY: int = 0
    nodeType: str = None
    actions: List[dict] = dict()  # todo: create a database action item
    transitions: Dict[str, dict] = dict()
    temp_on_receive: Optional[list] = list()

    def to_processed_node_element(self) -> ProcessedNodeElement:
        from inoft_vocal_engine.models.actions import get_appropriate_action_from_dict
        # The way the actions are stored in the database for the way the data is structured in the client
        # browser. In the browser, all the data (nodeId, actionId) is in the same object for simpler sending
        # to the database, where as in the database, we do not duplicate data if there is no need to.
        processed_actions: List[BaseAction] = list()
        for action in self.actions:
            action["nodeId"] = self.nodeId
            action["isNew"] = False

            action_instance = get_appropriate_action_from_dict(action_data_dict=action)
            if action_instance is not None:
                processed_actions.append(action_instance)

        processed_transitions: List[ProcessedTransitionCondition] = list()
        for transition_key_id, transition_values in self.transitions.items():
            processed_transitions.append(ProcessedTransitionCondition.from_database_item(
                DatabaseTransitionConditionItem(transitionId=transition_key_id, sourceNodeId=self.nodeId, **transition_values)
            ))

        return ProcessedNodeElement(
            nodeId=self.nodeId,
            nodeName=self.nodeName,
            actions=processed_actions,
            transitions=processed_transitions,
            temp_on_receive=self.temp_on_receive,
            nodeType=self.nodeType,
            positionX=self.positionX,
            positionY=self.positionY
        )




# invite : https://discordapp.com/oauth2/authorize?client_id=713363868226945054&scope=bot&permissions=8
# token :

from discord.ext import commands
from discord.ext.commands.context import Context
from discord.message import Message
from inoft_vocal_framework.platforms_handlers.discord import writing_func as wf
from inoft_vocal_framework.platforms_handlers.discord import DiscordStaticInfos


bot_client = commands.Bot(command_prefix=DiscordStaticInfos.COMMAND_PREFIX)
# If you have the error CERTIFICATE_VERIFY_FAILED, you have an issue with a missing SSL certificate on your computer.
# Go to this link : https://crt.sh/?id=2835394 and click on "Download Certificate: PEM" at the bottom right of the page,
# it will download a SSL Certificate, that you must install. PS : The certificate is only valid until 2038 !

private_state = False
m_type = 'launch'

@bot_client.event
async def on_ready():
    print('Connected')

@bot_client.command()
async def ping(context: Context):
    await context.send(f'Pong! {round(bot_client.latency*1000)} ms')

@bot_client.command()
async def private(context: Context):
    # user grabbing by id
    user = bot_client.get_user(context.message.author.id)
    # dm sending by user (id)
    await user.send('Voulez vous parler dans un channel privé ou non ?')

@bot_client.command()
async def launch(context: Context):
    await wf.send_poll('Voulez vous jouer en textuel privé ou public ?', context, 'num', m_type, 2)

@bot_client.command()
async def testpoll(context: Context):
    await wf.send_poll("poll test", context, 'num', m_type, 5)
    print(context)

@bot_client.command()
async def testmess(context: Context):
    await wf.send_message(message_to_send="message test", destination_context=context)

@bot_client.event
async def on_reaction_add(reaction, user):
    author = reaction.message.author

    if user == author:
        return
    else:
        await wf.emojis_reaction_response(m_type, reaction, user)

@bot_client.event
async def on_message(message: Message):
    prefix_debug = message.content[:1]
    if message.author == bot_client.user:
        return
    elif message.content[:1] != DiscordStaticInfos.COMMAND_PREFIX:

        # user grabbing by id
        user = bot_client.get_user(message.author.id)
        from inoft_vocal_framework.platforms_handlers.discord.handler_input import DiscordHandlerInput
        DiscordHandlerInput.HANDLER_FUNCTION_TO_USE(event=message, context=None)

        """
        print(f"Channel id beach {message.channel.id} & type : {type(message.channel.id)}")
        print(f"user id ea {message.author.id} & type : {type(message.author.id)}")
        print(f"Channel id beach {message.channel} & type : {type(message.channel)}")
        print(f"user id ea {message.author} & type : {type(message.author)}")

        data = {
            "author_name": message.author.name,
            "author_id": message.author.id,
            "message": message.content,
            "time": message.created_at
        }

        print(data)

        # public channel sending
        await message.channel.send(data)
        # sending private response to a spécific person whom we grabbed his id before
        await user.send(data)
        """

    await bot_client.process_commands(message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Automate Audacity via mod-script-pipe.
Pipe Client may be used as a command-line script to send commands to
Audacity via the mod-script-pipe interface, or loaded as a module.
Requires Python 2.7 or later. Python 3 strongly recommended.
======================
Command Line Interface
======================
    usage: pipeclient.py [-h] [-t] [-s ] [-d]
Arguments
---------
    -h,--help: optional
        show short help and exit
    -t, --timeout: float, optional
        timeout for reply in seconds (default: 10)
    -s, --show-time: bool, optional
        show command execution time (default: True)
    -d, --docs: optional
        show this documentation and exit
Example
-------
    $ python3 pipeclient.py -t 20 -s False
    Launches command line interface with 20 second time-out for
    returned message, and don't show the execution time.
    When prompted, enter the command to send (not quoted), or 'Q' to quit.
    $ Enter command or 'Q' to quit: GetInfo: Type=Tracks Format=LISP
============
Module Usage
============
Note that on a critical error (such as broken pipe), the module just exits.
If a more graceful shutdown is required, replace the sys.exit()'s with
exceptions.
Example
-------
    # Import the module:
    >>> import pipeclient
    # Create a client instance:
    >>> client = pipeclient.PipeClient()
    # Send a command:
    >>> client.write("Command", timer=True)
    # Read the last reply:
    >>> print(client.read())
See Also
--------
PipeClient.write : Write a command to _write_pipe.
PipeClient.read : Read Audacity's reply from pipe.
Copyright Steve Daulton 2018
Released under terms of the GNU General Public License version 2:
<http://www.gnu.org/licenses/old-licenses/gpl-2.0.html />
"""

import os
import sys
import threading
import time
import errno
import argparse


if sys.version_info[0] < 3 and sys.version_info[1] < 7:
    sys.exit('PipeClient Error: Python 2.7 or later required')

# Platform specific constants
if sys.platform == 'win32':
    WRITE_NAME = '\\\\.\\pipe\\ToSrvPipe'
    READ_NAME = '\\\\.\\pipe\\FromSrvPipe'
    EOL = '\r\n\0'
else:
    # Linux or Mac
    PIPE_BASE = '/tmp/audacity_script_pipe.'
    WRITE_NAME = PIPE_BASE + 'to.' + str(os.getuid())
    READ_NAME = PIPE_BASE + 'from.' + str(os.getuid())
    EOL = '\n'


class PipeClient(object):
    """Write / read client access to Audacity via named pipes.
    Normally there should be just one instance of this class. If
    more instances are created, they all share the same state.
    __init__ calls _write_thread_start() and _read_thread_start() on
    first instantiation.
    Parameters
    ----------
        None
    Attributes
    ----------
        reader_pipe_broken : event object
            Set if pipe reader fails. Audacity may have crashed
        reply_ready : event object
            flag cleared when command sent and set when response received
        timer : bool
            When true, time the command execution (default False)
        reply : string
            message received when Audacity completes the command
    See Also
    --------
    write : Write a command to _write_pipe.
    read : Read Audacity's reply from pipe.
    """

    reader_pipe_broken = threading.Event()
    reply_ready = threading.Event()

    _shared_state = {}

    def __new__(cls, *p, **k):
        self = object.__new__(cls, *p, **k)
        self.__dict__ = cls._shared_state
        return self

    def __init__(self):
        self.timer = False
        self._start_time = 0
        self._write_pipe = None
        self.reply = ''
        if not self._write_pipe:
            self._write_thread_start()
        self._read_thread_start()

    def _write_thread_start(self):
        """Start _write_pipe thread"""
        # Pipe is opened in a new thread so that we don't
        # freeze if Audacity is not running.
        write_thread = threading.Thread(target=self._write_pipe_open)
        write_thread.daemon = True
        write_thread.start()
        # Allow a little time for connection to be made.
        time.sleep(0.1)
        if not self._write_pipe:
            sys.exit('PipeClientError: Write pipe cannot be opened.')

    def _write_pipe_open(self):
        """Open _write_pipe."""
        self._write_pipe = open(WRITE_NAME, 'w')

    def _read_thread_start(self):
        """Start read_pipe thread."""
        read_thread = threading.Thread(target=self._reader)
        read_thread.daemon = True
        read_thread.start()

    def write(self, command, timer=False):
        """Write a command to _write_pipe.
        Parameters
        ----------
            command : string
                The command to send to Audacity
            timer : bool, optional
                If true, time the execution of the command
        Example
        -------
            write("GetInfo: Type=Labels", timer=True):
        """
        self.timer = timer
        print('Sending command:', command)
        self._write_pipe.write(command + EOL)
        # Check that read pipe is alive
        if PipeClient.reader_pipe_broken.isSet():
            sys.exit('PipeClient: Read-pipe error.')
        try:
            self._write_pipe.flush()
            if self.timer:
                self._start_time = time.time()
            self.reply = ''
            PipeClient.reply_ready.clear()
        except IOError as err:
            if err.errno == errno.EPIPE:
                sys.exit('PipeClient: Write-pipe error.')
            else:
                raise

    def _reader(self):
        """Read FIFO in worker thread."""
        # Thread will wait at this read until it connects.
        # Connection should occur as soon as _write_pipe has connected.
        read_pipe = open(READ_NAME, 'r')
        message = ''
        while True:
            line = read_pipe.readline()
            # Stop timer as soon as we get first line of response.
            stop_time = time.time()
            while line != '\n':
                message += line
                line = read_pipe.readline()
                if line == '':
                    # No data in read_pipe indicates that the pipe is broken
                    # (Audacity may have crashed).
                    PipeClient.reader_pipe_broken.set()
            if self.timer:
                xtime = (stop_time - self._start_time) * 1000
                message += 'Execution time: {0:.2f}ms'.format(xtime)
            self.reply = message
            PipeClient.reply_ready.set()
            message = ''
        read_pipe.close()

    def read(self):
        """Read Audacity's reply from pipe.
        Returns
        -------
        string
            The reply from the last command sent to Audacity, or null string
            if reply not received. Null string usually indicates that Audacity
            is still processing the last command.
        """
        if not PipeClient.reply_ready.isSet():
            return ''
        else:
            return self.reply


def bool_from_string(strval):
    """Return boolean value from string"""
    if strval.lower() in ('true', 't', '1', 'yes', 'y'):
        return True
    elif strval.lower() in ('false', 'f', '0', 'no', 'n'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def main():
    """Interactive command-line for PipeClient"""

    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--timeout', type=float, metavar='', default=10,
                        help="timeout for reply in seconds (default: 10")
    parser.add_argument('-s', '--show-time', metavar='True/False',
                        nargs='?', type=bool_from_string,
                        const='t', default='t', dest='show',
                        help='show command execution time (default: True)')
    parser.add_argument('-d', '--docs', action='store_true',
                        help='show documentation and exit')
    args = parser.parse_args()

    if args.docs:
        print(__doc__)
        sys.exit(0)

    client = PipeClient()
    while True:
        reply = ''
        if sys.version_info[0] < 3:
            #pylint: disable=undefined-variable
            message = raw_input("\nEnter command or 'Q' to quit: ")
        else:
            message = input( #pylint: disable=bad-builtin
                "\nEnter command or 'Q' to quit: ")
        start = time.time()
        if message.upper() == 'Q':
            sys.exit(0)
        elif message == '':
            pass
        else:
            client.write(message, timer=args.show)
            while reply == '':
                time.sleep(0.1)  # allow time for reply
                if time.time() - start > args.timeout:
                    reply = 'PipeClient: Reply timed-out.'
                else:
                    reply = client.read()
            print(reply)


if __name__ == '__main__':
    main()



from typing import Optional

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.databases.dynamodb.audio_editor_projects_dynamodb_client import AudioEditorProjectsDynamoDbClient
from inoft_vocal_engine.databases.dynamodb.projects_diagrams_data_dynamodb_client import ProjectsDiagramsDataDynamoDbClient
from inoft_vocal_engine.databases.dynamodb.projects_text_contents_dynamodb_client import \
    ProjectsTextContentsDynamoDbClient, ProjectsTextContentsDynamoDBNewTableClient
from inoft_vocal_engine.databases.dynamodb.team_organization_projects_dynamodb_client import TeamOrganizationProjectsDynamoDbClient
from inoft_vocal_engine.databases.dynamodb.vocal_apps_model_schemas_dynamodb_client import VocalAppsModelSchemasDynamoDBClient
from inoft_vocal_engine.vocal_apps_model_schemas.models import InoftVocalEngineModelSchema


class ProjectResources:
    _project_text_contents_table = None

    _audio_editor_projects_dynamodb_client = None
    _project_text_contents_dynamodb_client = None
    _project_diagrams_data_dynamodb_client = None
    _vocal_apps_model_schemas_dynamodb_client = None
    _team_organization_projects_dynamodb_client = None

    _audio_editor_projects_dynamodb_table_name = "inoft-vocal-engine_accounts-projects-instances_audio-editor-projects-data"
    _project_text_contents_dynamodb_table_name = "inoft-vocal-engine_accounts-projects-instances_text-contents-data"
    _project_diagrams_data_dynamodb_table_name = "inoft-vocal-engine_accounts-projects-instances_diagrams-nodes-data"
    _vocal_apps_model_schemas_dynamodb_table_name = "inoft-vocal-engine_accounts-projects_vocal-apps-model-schemas"
    _team_organization_projects_dynamodb_table_name = "inoft-vocal-engine_accounts-projects-instances_team-organization-projects-data"

    def __init__(self, engine_resources: EngineResources, account_resources: AccountResources,
                 project_url: str, project_owner_account_username: str, project_owner_account_id: str,
                 account_project_id: Optional[str] = None):

        self.engine_resources = engine_resources
        self.account_resources = account_resources
        self.project_url = project_url
        self.project_owner_account_username = project_owner_account_username
        self.project_owner_account_id = project_owner_account_id
        self._account_project_id = account_project_id
        # We have the possibility to set the account_project_id from an argument, because in the
        # auth_workflow, we need to get the account_project_id before we create the ProjectResources
        # instance, because the account_project_id is used in the permissions format.

    @property
    def account_project_id(self) -> str:
        if self._account_project_id is None:
            self._account_project_id = self.engine_resources.accounts_data_dynamodb_client.get_account_project_id_by_project_url(
                account_username=self.account_resources.account_username, project_url=self.project_url
            )
        return self._account_project_id

    """
    @property
    def _project_text_contents_table(self) -> StructNoSQL.BaseTable:
        if self._project_text_contents_table is None:
            self._project_text_contents_table = ProjectsTextContentsDynamoDBNewTableClient(
                table_name=self._project_text_contents_dynamodb_table_name, region_name="eu-west-2"
            )
        return self._audio_editor_projects_dynamodb_client
    """

    @property
    def audio_editor_projects_dynamodb_client(self) -> AudioEditorProjectsDynamoDbClient:
        if self._audio_editor_projects_dynamodb_client is None:
            self._audio_editor_projects_dynamodb_client = AudioEditorProjectsDynamoDbClient(
                table_name=self._audio_editor_projects_dynamodb_table_name, region_name="eu-west-2"
            )
        return self._audio_editor_projects_dynamodb_client

    @property
    def project_text_contents_dynamodb_client(self) -> ProjectsTextContentsDynamoDbClient:
        if self._project_text_contents_dynamodb_client is None:
            self._project_text_contents_dynamodb_client = ProjectsTextContentsDynamoDbClient(
                table_name=self._project_text_contents_dynamodb_table_name, region_name="eu-west-2"
            )
        return self._project_text_contents_dynamodb_client

    @property
    def project_diagrams_data_dynamodb_client(self) -> ProjectsDiagramsDataDynamoDbClient:
        if self._project_diagrams_data_dynamodb_client is None:
            self._project_diagrams_data_dynamodb_client = ProjectsDiagramsDataDynamoDbClient(
                table_name=self._project_diagrams_data_dynamodb_table_name, region_name="eu-west-2"
            )
        return self._project_diagrams_data_dynamodb_client

    @property
    def vocal_apps_model_schemas_dynamodb_client(self) -> VocalAppsModelSchemasDynamoDBClient:
        if self._vocal_apps_model_schemas_dynamodb_client is None:
            self._vocal_apps_model_schemas_dynamodb_client = VocalAppsModelSchemasDynamoDBClient(
                table_name=self._vocal_apps_model_schemas_dynamodb_table_name, region_name="eu-west-2"
            )
        return self._vocal_apps_model_schemas_dynamodb_client

    @property
    def team_organization_projects_dynamodb_client(self) -> TeamOrganizationProjectsDynamoDbClient:
        if self._team_organization_projects_dynamodb_client is None:
            self._team_organization_projects_dynamodb_client = TeamOrganizationProjectsDynamoDbClient(
                table_name=self._team_organization_projects_dynamodb_table_name, region_name="eu-west-2"
            )
        return self._team_organization_projects_dynamodb_client

    def required_template_kwargs(self) -> dict:
        return {
            "project_owner_account_username": self.project_owner_account_username,
            "project_url": self.project_url
        }

    def deploy_lambda(self):
        from inoft_vocal_engine.cloud_providers.aws.deploy import DeployHandler
        from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path
        DeployHandler().deploy(lambda_files_root_folderpath=get_inoft_vocal_engine_root_path(), stage_name="prod",
                               bucket_name="inoft-vocal-engine-web-test", bucket_region_name="eu-west-3",
                               lambda_name="inoft-vocal-engine-web-interface",
                               lambda_handler="inoft_vocal_engine.web_interface.app.lambda_handler", runtime="python3.7")

    def deploy_project(self):
        from inoft_vocal_engine.cloud_providers.aws.deploy import DeployHandler
        from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path
        DeployHandler().deploy(lambda_files_root_folderpath=get_inoft_vocal_engine_root_path(), stage_name="prod",
                               bucket_name="inoft-vocal-engine-web-test", bucket_region_name="eu-west-3",
                               lambda_name="inoft-vocal-engine-web-interface",
                               lambda_handler="inoft_vocal_engine.web_interface.app.lambda_handler", runtime="python3.7")


if __name__ == "__main__":
    engine_resources = EngineResources()
    account_resources = AccountResources(engine_resources=engine_resources)
    project_resources = ProjectResources(engine_resources=engine_resources, account_resources=account_resources,
                                         project_url="anvers1944", project_owner_account_username="robinsonlabourdette",
                                         project_owner_account_id="b1fe5939-032b-462d-92e0-a942cd445096")
    project_resources.deploy_lambda()
    # print(resources.project_diagrams_data_dynamodb_client.query_nodes_by_accountProjectId(accountProjectId="5aedb52e-3431-42dc-bf0b-ccd2844d2607"))
    """print(resources.vocal_apps_model_schemas_dynamodb_client.set_update_one_intent(
        accountProjectId=resources.account_project_id,
        intent_model=InoftVocalEngineModelSchema.IntentModel(name="test", samples=["j'aime les test", "yolo"])))"""
    """print(resources.vocal_apps_model_schemas_dynamodb_client.put_model_schema(
        account_project_id=resources.account_project_id,
        model_schema=InoftVocalEngineModelSchema(
        invocationName="test", LanguageModel="fr-FR",
        alexaSpecific={
            "modelConfiguration": {
                "fallbackIntentSensitivity": {
                    "level": "LOW"
                }
            }
        }
    )))"""
